from lasagna import (
    known_models,
    build_simple_agent,
)

from lasagna.tui import (
    tui_input_loop,
)

from typing import List, Callable

from dotenv import load_dotenv; load_dotenv()

import asyncio
import aiohttp
import os
from datetime import datetime


MODEL_BINDER = known_models.BIND_OPENAI_gpt_4o_mini()


async def perform_research(query: str) -> str:
    """
    Use this tool to do internet research, for both:
    1. to do research about things you don't know, and
    2. to get information about recent news or events.
    :param: query: str: the query string to use for the internet search
    """
    url = 'https://api.tavily.com/search'
    payload = {
      "api_key": os.environ['TAVILY_API_KEY'],
      "query": query,
      "search_depth": "basic",   # 'basic' or 'advanced'
      "include_answer": False,
      "include_images": False,
      "include_raw_content": False,
      "max_results": 5,
    }
    async with aiohttp.ClientSession(raise_for_status=True) as session:
        async with session.post(url, json=payload) as request:
            response = await request.json()
    results = [
        f"Source: {obj['url']}\n\nContent: {obj['content']}"
        for obj in response['results']
    ]
    return '\n\n'.join(results)


async def main() -> None:
    today = datetime.now().strftime('%B %d, %Y')
    system_prompt = f"You are a grumpy research agent. Today is {today}."
    tools: List[Callable] = [
        perform_research,
    ]
    my_agent = build_simple_agent(name = 'agent', tools = tools)
    my_bound_agent = MODEL_BINDER(my_agent)
    await tui_input_loop(my_bound_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())



"""
This module is the Lasagna adapter for the Ollama server.

For more information about Ollama, see:
 - https://ollama.com/
"""

from .types import (
    Cost,
    EventPayload,
    MessageToolCall,
    ModelSpec,
    Message,
    Media,
    EventCallback,
    ToolCall,
    Model,
    ExtractionType,
)

from .util import (
    convert_to_image_base64,
    exponential_backoff_retry_delays,
    get_name,
    recursive_hash,
)

from .tools_util import (
    convert_to_json_schema,
    extract_tool_result_as_sting,
    get_tool_params,
    handle_tools,
    build_tool_response_message,
)

from .pydantic_util import ensure_pydantic_model, build_and_validate

from openai.lib._pydantic import to_strict_json_schema

from typing import (
    List, Callable, AsyncIterator, Any, Type,
    Tuple, Dict, Union,
)

import os
import asyncio
import httpx
import copy
import json

import logging

_LOG = logging.getLogger(__name__)


def _convert_to_ollama_tool(tool: Callable) -> Dict:
    description, params = get_tool_params(tool)
    return {
        'type': 'function',
        'function': {
            'name': get_name(tool),
            'description': description,
            'parameters': convert_to_json_schema(params),
        },
    }


def _convert_to_ollama_tools(tools: List[Callable]) -> Union[None, List[Dict]]:
    if len(tools) == 0:
        return None
    specs = [_convert_to_ollama_tool(tool) for tool in tools]
    return specs


def _log_dumps(val: Any) -> str:
    if isinstance(val, dict):
        return json.dumps(val)
    else:
        return str(val)


async def _convert_to_ollama_media(media: List[Media]) -> Dict:
    res: Dict = {
        'images': [],
    }
    for m in media:
        assert m['type'] == 'image'
        mimetype, data = await convert_to_image_base64(m['image'])
        assert mimetype
        res['images'].append(data)
    return res


def _convert_to_ollama_tool_calls(tools: List[ToolCall]) -> List:
    res: List = []
    for t in tools:
        assert t['call_type'] == 'function'
        res.append({
            # t['call_id'] NOT USED! Ollama doesn't do that sort of thing.
            'function': {
                'name': t['function']['name'],
                'arguments': json.loads(t['function']['arguments']),
            },
        })
    return res


async def _convert_to_ollama_messages(messages: List[Message]) -> List[Dict]:
    res: List[Dict] = []
    map = {
        'system': 'system',
        'human': 'user',
        'ai': 'assistant',
    }
    prev_tool_call_map = {}
    for m in messages:
        if m['role'] == 'system' or m['role'] == 'human' or m['role'] == 'ai':  # <-- not using boolean 'in' to make mypy happy
            media = {}
            if 'media' in m and m['media']:
                media = await _convert_to_ollama_media(m['media'])
            res.append({
                'role': map[m['role']],
                'content': m.get('text', ''),
                **media,
            })
        elif m['role'] == 'tool_call':
            res.append({
                'role': 'assistant',
                'content': '',
                'tool_calls': _convert_to_ollama_tool_calls(m['tools']),
            })
            prev_tool_call_map = {
                t['call_id']: t['function']['name']
                for t in m['tools']
            }
        elif m['role'] == 'tool_res':
            for t in m['tools']:
                # t['is_error'] IS NOT USED!
                res.append({
                    'role': 'tool',
                    'content': extract_tool_result_as_sting(t),
                    'name': prev_tool_call_map.get(t['call_id'], 'unknown'),
                })
        else:
            raise RuntimeError(f"unreachable: {m['role']}")
    return res


async def _event_stream(
    url: str,
    payload: Dict,
    timeout_seconds: float,
) -> AsyncIterator[Dict]:
    timeout = httpx.Timeout(timeout_seconds, connect=2.0)  # 2s timeout on connect. `timeout_seconds` elsewhere (read/write/pool).
    async with httpx.AsyncClient(timeout=timeout) as client:
        async with client.stream('POST', url, json=payload) as r:
            if r.status_code != 200:
                error_text = json.loads(await r.aread())['error']
                raise RuntimeError(f'Ollama error: {error_text}')
            async for line in r.aiter_lines():
                rec = json.loads(line)
                if rec.get('error'):
                    error_text = rec['error']
                    raise RuntimeError(f'Ollama error: {error_text}')
                yield rec


def _set_cost_raw(message: Message, raw: List[Dict]) -> Message:
    cost: Cost = {
        'input_tokens': None,
        'output_tokens': None,
        'total_tokens': None,
    }
    for event in raw:
        if 'prompt_eval_count' in event:
            cost['input_tokens'] = (cost['input_tokens'] or 0) + event['prompt_eval_count']
        if 'eval_count' in event:
            cost['output_tokens'] = (cost['output_tokens'] or 0) + event['eval_count']
    cost['total_tokens'] = (cost['input_tokens'] or 0) + (cost['output_tokens'] or 0)
    new_message: Message = copy.copy(message)
    if (cost['total_tokens'] or 0) > 0:
        new_message['cost'] = cost
    new_message['raw'] = raw
    return new_message


async def _process_stream(
    stream: AsyncIterator[Dict],
    event_callback: EventCallback,
) -> List[Message]:
    raw: List[Dict] = []
    content: List[str] = []
    tools: List[ToolCall] = []
    async for event in stream:
        raw.append(event)
        if 'message' in event:
            m = event['message']
            if 'content' in m:
                c = m['content']
                if c:
                    assert isinstance(c, str)
                    await event_callback(('ai', 'text_event', c))
                    content.append(c)
            if 'tool_calls' in m:
                for i, t in enumerate(m['tool_calls']):
                    assert 'function' in t
                    f = t['function']
                    name = f['name']
                    args = f['arguments']  # is a dict
                    args_str = json.dumps(args)
                    tool_call: ToolCall = {
                        'call_id': f'call_{i}',  # <-- Ollama doesn't do the `call_id` thing, so we invent a call_id
                        'call_type': 'function',
                        'function': {
                            'name': name,
                            'arguments': args_str,
                        },
                    }
                    await event_callback(('tool_call', 'text_event', f'{name}({args_str})\n'))
                    await event_callback(('tool_call', 'tool_call_event', tool_call))
                    tools.append(tool_call)
    messages: List[Message] = []
    if content:
        messages.append({
            'role': 'ai',
            'text': ''.join(content),
        })
    if tools:
        messages.append({
            'role': 'tool_call',
            'tools': tools,
        })
    if len(messages) > 0:
        messages[-1] = _set_cost_raw(messages[-1], raw)
    return messages


def _wrap_event_callback_convert_ai_text_to_tool_call_text(
    wrapped: EventCallback,
) -> EventCallback:
    async def wrapper(event: EventPayload) -> None:
        if event[0] == 'ai' and event[1] == 'text_event':
            await wrapped(('tool_call', 'text_event', event[2]))
        else:
            await wrapped(event)

    return wrapper


def _get_ollama_format_for_structured_output(
    extraction_type: Type[ExtractionType],
) -> Dict:
    format: Dict = to_strict_json_schema(ensure_pydantic_model(extraction_type))

    docstr = getattr(extraction_type, '__doc__', None)
    if docstr:
        format['description'] = docstr

    return format


class LasagnaOllama(Model):
    def __init__(self, model: str, **model_kwargs: Any):
        self.model = model
        self.model_kwargs = copy.deepcopy(model_kwargs)
        self.n_retries: int = self.model_kwargs.get('retries', 3)
        if not isinstance(self.n_retries, int) or self.n_retries < 0:
            raise ValueError(f"model_kwargs['retries'] must be a non-negative integer (got {self.n_retries})")
        self.model_spec: ModelSpec = {
            'provider': 'ollama',
            'model': self.model,
            'model_kwargs': self.model_kwargs,
        }

        self.base_url: str = self.model_kwargs.get('base_url', os.environ.get('OLLAMA_BASE_URL', 'http://127.0.0.1:11434'))
        assert isinstance(self.base_url, str)

        self.keep_alive: str = self.model_kwargs.get('keep_alive', '5m')
        assert isinstance(self.keep_alive, str)

        self.timeout_seconds = float(self.model_kwargs.get('timeout_seconds', 120.0))
        if self.timeout_seconds < 0.0:
            raise ValueError(f"model_kwargs['timeout_seconds'] must be a non-negative float (got {self.timeout_seconds})")

        self.payload_options = copy.deepcopy(self.model_kwargs)
        for key_to_remove in ['retries', 'base_url', 'keep_alive', 'timeout_seconds']:
            if key_to_remove in self.payload_options:
                del self.payload_options[key_to_remove]

    def config_hash(self) -> str:
        return recursive_hash(None, self.model_spec)

    async def _run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[None, List[Dict]],
        force_tool: bool,
        format: Union[None, Dict],
    ) -> List[Message]:
        stream = True

        if tools_spec and format:
            raise ValueError("Oops! You cannot do both tool-use and structured output at the same time!")

        if tools_spec:
            stream = False   # Ollama does not yet support streaming tool responses.
            if not force_tool:
                raise ValueError("Oops! Ollama currently does not support *optional* tool use. Thus, if you pass tools, you must also pass `force_tool=True` to show that your intended use matches Ollama's behavior.")
        else:
            if force_tool:
                raise ValueError("Oops! You cannot force tools that are not specified!")

        ollama_messages = await _convert_to_ollama_messages(messages)

        _LOG.info(f"Invoking {self.model} with:\n  messages: {_log_dumps(ollama_messages)}\n  tools: {_log_dumps(tools_spec)}")

        url = f'{self.base_url}/api/chat'

        payload = {
            'model': self.model,
            'messages': ollama_messages,
            'stream': stream,
            **({'tools': tools_spec} if tools_spec else {}),
            **({'format': format} if format else {}),
            'options': self.payload_options,
            'keep_alive': self.keep_alive,
        }

        event_stream = _event_stream(url, payload, self.timeout_seconds)
        new_messages = await _process_stream(event_stream, event_callback)

        _LOG.info(f"Finished {self.model}")

        return new_messages

    async def _retrying_run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[None, List[Dict]],
        force_tool: bool,
        format: Union[None, Dict],
    ) -> List[Message]:
        last_error: Union[Exception, None] = None
        assert self.n_retries + 1 > 0   # <-- we know this is true from the check in __init__
        for delay_on_error in exponential_backoff_retry_delays(self.n_retries + 1):
            try:
                await event_callback(('transaction', 'start', ('ollama', self.model)))
                try:
                    new_messages = await self._run_once(
                        event_callback = event_callback,
                        messages = messages,
                        tools_spec = tools_spec,
                        force_tool = force_tool,
                        format = format,
                    )
                except:
                    await event_callback(('transaction', 'rollback', None))
                    raise
                await event_callback(('transaction', 'commit', None))
                return new_messages
            except Exception as e:
                # Some errors should be retried, some should not. Below
                # is the logic to decide when to retry vs when to not.
                # It's likely this will change as we get more usage and see
                # where Ollama tends to fail, and when we know more what is
                # recoverable vs not.
                last_error = e
                if isinstance(e, httpx.HTTPError):
                    # Network connection error. We can retry this.
                    pass
                else:
                    # Some other error that we don't know about. Let's bail.
                    raise
                if delay_on_error > 0.0:
                    _LOG.warning(f"Got a maybe-recoverable error (will retry in {delay_on_error:.2f} seconds): {e}")
                    await asyncio.sleep(delay_on_error)
        assert last_error is not None   # <-- we know this is true because `n_retries + 1 > 0`
        raise last_error

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        messages = [*messages]  # shallow copy
        new_messages: List[Message] = []
        tools_spec = _convert_to_ollama_tools(tools)
        tools_map = {get_name(tool): tool for tool in tools}
        for _ in range(max_tool_iters):
            new_messages_here = await self._retrying_run_once(
                event_callback = event_callback,
                messages       = messages,
                tools_spec     = tools_spec,
                force_tool     = force_tool,
                format         = None,
            )
            tools_results = await handle_tools(
                prev_messages = messages,
                new_messages = new_messages_here,
                tools_map = tools_map,
                event_callback = event_callback,
                model_spec = self.model_spec,
            )
            new_messages.extend(new_messages_here)
            messages.extend(new_messages_here)
            if tools_results is None:
                break
            for tool_result in tools_results:
                await event_callback(('tool_res', 'tool_res_event', tool_result))
            tool_response_message = build_tool_response_message(tools_results)
            new_messages.append(tool_response_message)
            messages.append(tool_response_message)
        return new_messages

    async def extract(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        extraction_type: Type[ExtractionType],
    ) -> Tuple[Message, ExtractionType]:
        format = _get_ollama_format_for_structured_output(extraction_type)

        new_messages = await self._retrying_run_once(
            event_callback = _wrap_event_callback_convert_ai_text_to_tool_call_text(event_callback),
            messages       = messages,
            tools_spec     = None,
            force_tool     = False,
            format         = format,
        )

        assert len(new_messages) == 1
        new_message = new_messages[0]

        assert new_message['role'] == 'ai'  # Ollama generates structured output just like it does normal text, so at this point it is just text.

        arguments = new_message['text'] or '{}'

        tool_call: ToolCall = {
            'call_id': 'extraction_tool_call',
            'call_type': 'function',
            'function': {
                'name': get_name(extraction_type),
                'arguments': arguments,
            },
        }

        tool_message: MessageToolCall = {
            'role': 'tool_call',
            'tools': [tool_call],
        }
        if 'cost' in new_message:
            tool_message['cost'] = new_message['cost']
        if 'raw' in new_message:
            tool_message['raw'] = new_message['raw']

        await event_callback(('tool_call', 'tool_call_event', tool_call))

        parsed = json.loads(arguments)
        result = build_and_validate(extraction_type, parsed)

        return tool_message, result



from .types import (
    AgentRecord,
    AgentCallable,
    ProviderRecord,
    ModelFactory,
    ModelRecord,
)

from typing import Dict, List


AGENTS: Dict[str, AgentRecord] = {}

PROVIDERS: Dict[str, ProviderRecord] = {}


def register_agent(
    key: str,
    name: str,
    runner: AgentCallable,
) -> None:
    if key in AGENTS:
        raise RuntimeError(f"An agent with this key is already registered: {key}")
    AGENTS[key] = {
        'name': name,
        'runner': runner,
    }


def register_provider(
    key: str,
    name: str,
    factory: ModelFactory,
    models: List[ModelRecord],
) -> None:
    if key in PROVIDERS:
        raise RuntimeError(f"A model provider with this key is already registered: {key}")
    PROVIDERS[key] = {
        'name': name,
        'factory': factory,
        'models': models,
    }



from .registrar import AGENTS, PROVIDERS

from .util import get_name

from .types import (
    AgentSpec,
    AgentRun,
    AgentCallable,
    ModelFactory,
    EventCallback,
)

from .known_providers import attempt_load_known_providers

from typing import List


async def run(
    agent_spec: AgentSpec,
    event_callback: EventCallback,
    prev_runs: List[AgentRun],
) -> AgentRun:
    agent: AgentCallable
    agent_name: str
    if isinstance(agent_spec['agent'], str):
        agent_name = agent_spec['agent']
        agent = AGENTS[agent_name]['runner']
    else:
        agent = agent_spec['agent']
        agent_name = get_name(agent)

    model_name = agent_spec['model'] \
        if isinstance(agent_spec['model'], str) \
        else agent_spec['model']['formal_name']

    model_kwargs = agent_spec.get('model_kwargs', None)
    if model_kwargs is None:
        model_kwargs = {}

    model_factory: ModelFactory
    provider_str: str
    if isinstance(agent_spec['provider'], str):
        provider_str = agent_spec['provider']
        if provider_str not in PROVIDERS:
            attempt_load_known_providers(provider_str)
        model_factory = PROVIDERS[provider_str]['factory']
        model = model_factory(model=model_name, **model_kwargs)
    else:
        model_factory = agent_spec['provider']
        model = model_factory(model=model_name, **model_kwargs)
        provider_str = get_name(model.__class__)

    await event_callback(('agent', 'start', agent_name))

    agent_run = await agent(
        model,
        event_callback,
        prev_runs,
    )

    if 'provider' not in agent_run:
        agent_run['provider'] = provider_str
    if 'model' not in agent_run:
        agent_run['model'] = model_name
    if 'model_kwargs' not in agent_run and 'model_kwargs' in agent_spec:
        agent_run['model_kwargs'] = model_kwargs

    await event_callback(('agent', 'end', agent_run))

    return agent_run



from __future__ import annotations
import abc

from typing import (
    TypedDict, Dict, List, Any, Optional, Literal,
    Callable, Tuple, Awaitable, Union, Protocol,
    TypeVar, Type,
)

from typing_extensions import NotRequired


class ToolCallFunction(TypedDict):
    name: str
    arguments: str   # <-- JSON-encoded arguments (kept as a string for reproducibility reasons)


class ToolCall(TypedDict):
    call_id: str
    call_type: Literal['function']   # <-- functions are the only type of "tool use" right now!
    function: ToolCallFunction


class ToolResultBase(TypedDict):
    call_id: str
    is_error: NotRequired[bool]


class ToolResultAny(ToolResultBase):
    type: Literal['any']
    result: Any


class ToolResultLayeredAgent(ToolResultBase):
    type: Literal['layered_agent']
    run: AgentRun


ToolResult = Union[
    ToolResultAny,
    ToolResultLayeredAgent,
]


class Cost(TypedDict):
    input_tokens: Optional[int]
    output_tokens: Optional[int]
    total_tokens: Optional[int]


class MediaImage(TypedDict):
    type: Literal['image']
    image: str   # <-- either a local filepath or a remote URL


Media = MediaImage   # note: this will become a Union[MediaImage, ...] in the future as needed


class MessageBase(TypedDict):
    cost: NotRequired[Optional[Cost]]  # <-- cost of generating this message (for messages generated by an external service)
    raw: NotRequired[Any]              # <-- the raw payload(s) from the API (before being converted our our schema; only for messages generated by an external service)


class MessageContent(MessageBase):
    role: Literal['system', 'human', 'ai']
    text: Optional[str]
    media: NotRequired[List[Media]]


class MessageToolCall(MessageBase):
    role: Literal['tool_call']
    tools: List[ToolCall]


class MessageToolResult(MessageBase):
    role: Literal['tool_res']
    tools: List[ToolResult]


Message = Union[
    MessageContent,
    MessageToolCall,
    MessageToolResult,
]


class AgentRunBase(TypedDict):
    agent: str
    provider: NotRequired[str]
    model: NotRequired[str]
    model_kwargs: NotRequired[Dict[str, Any]]


class AgentRunMessageList(AgentRunBase):
    type: Literal['messages']
    messages: List[Message]


class AgentRunParallel(AgentRunBase):
    type: Literal['parallel']
    runs: List[AgentRun]


class AgentRunChained(AgentRunBase):
    type: Literal['chain']
    runs: List[AgentRun]


class AgentRunExtraction(AgentRunBase):
    type: Literal['extraction']
    messages: List[Message]
    result: Any    # ideally, it's `ExtractionType`, and we'd inherit from Generic[ExtractionType], but that's not supported until python3.11


AgentRun = Union[
    AgentRunMessageList,
    AgentRunParallel,
    AgentRunChained,
    AgentRunExtraction,
]


EventPayload = Union[
    Tuple[Literal['human'],       Literal['echo_event'],      MessageContent],

    Tuple[Literal['ai'],          Literal['text_event'],      str],

    Tuple[Literal['tool_call'],   Literal['text_event'],      str],
    Tuple[Literal['tool_call'],   Literal['tool_call_event'], ToolCall],

    Tuple[Literal['tool_res'],    Literal['tool_res_event'],  ToolResult],

    Tuple[Literal['progress'],    Literal['start'],           Tuple[str, str]],    # payload is `(key, details)`
    Tuple[Literal['progress'],    Literal['update'],          Tuple[str, float]],  # payload is `(key, progress_0_to_1)`
    Tuple[Literal['progress'],    Literal['end'],             str],                # payload is `key`

    Tuple[Literal['transaction'], Literal['start'],           Tuple[str, str]],    # payload is `(provider, model)`
    Tuple[Literal['transaction'], Literal['rollback'],        None],
    Tuple[Literal['transaction'], Literal['commit'],          None],

    Tuple[Literal['agent'],       Literal['start'],           str],                # payload is `agent_name`
    Tuple[Literal['agent'],       Literal['end'],             AgentRun],
]


EventCallback = Callable[[EventPayload], Awaitable[None]]


ExtractionType = TypeVar('ExtractionType')  # <-- for generic programming


class Model(abc.ABC):
    """
    Interface for an AI model.

    In many cases this will be an LLM (when you're in pure-text use-cases). But
    this interface also supports AI models that operate on multimodal content,
    thus we'll use the phrase "AI model" to be more general.
    """

    @abc.abstractmethod
    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        """
        Generate one-or-more responses from the AI. More than one response
        is generated when tools are used. This method will respond to tool-
        use requests until the AI generates a non-tool response (up to
        `max_tool_iters`, then it halts).
        """
        pass

    @abc.abstractmethod
    async def extract(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        extraction_type: Type[ExtractionType],
    ) -> Tuple[Message, ExtractionType]:
        """
        Use the AI to extract structured output from the `messages`. The
        schema of the extracted data will follow `extraction_type`.
        """
        pass

    @abc.abstractmethod
    def config_hash(self) -> str:
        """
        Returns a hash of this models configuration. This likely contains
        the provider's name, model's name, and any model_kwargs. You can use
        any algorithm you'd like as long as it's stable from one Python version
        to the next; e.g., we recommend you use `util.recursive_hash()`.
        """
        pass


AgentCallable = Callable[[Model, EventCallback, List[AgentRun]], Awaitable[AgentRun]]

BoundAgentCallable = Callable[[EventCallback, List[AgentRun]], Awaitable[AgentRun]]


class AgentRecord(TypedDict):
    name: str
    runner: AgentCallable


class ModelRecord(TypedDict):
    formal_name: str
    display_name: str
    outdated: NotRequired[bool]  # <-- if true, there is another model which replaces this one


class ModelFactory(Protocol):
    def __call__(self, model: str, **model_kwargs: Any) -> Model: ...


class ProviderRecord(TypedDict):
    name: str
    factory: ModelFactory
    models: List[ModelRecord]


class ModelSpec(TypedDict):
    provider: Union[str, ModelFactory]
    model: Union[str, ModelRecord]
    model_kwargs: NotRequired[Dict[str, Any]]


class AgentSpec(ModelSpec):
    agent: Union[str, AgentCallable]


class ToolParam(TypedDict):
    name: str
    type: str
    description: NotRequired[str]
    optional: NotRequired[bool]  # default = False (i.e. "required")


class CacheEventPayload(TypedDict):
    delta_time: float   # <-- when this event arrived (as the number of second since the start of the agent's execution)
    event: EventPayload


CacheKey = str


class CacheRecord(TypedDict):
    events: List[CacheEventPayload]
    run: AgentRun


ImageMimeTypes = Literal['image/jpeg', 'image/png', 'image/gif', 'image/webp']



"""
This module is the Lasagna adapter for the Anthropic models.

For more information about the Anthropic models this adapter is for, see:
 - https://docs.anthropic.com/en/docs/welcome
"""

from .types import (
    ModelSpec,
    Message,
    MessageContent,
    MessageToolCall,
    MessageToolResult,
    EventCallback,
    EventPayload,
    ToolCall,
    Model,
    Cost,
    ExtractionType,
)

from .stream import (
    adup,
)

from .util import (
    convert_to_image_base64,
    exponential_backoff_retry_delays,
    get_name,
    recursive_hash,
)

from .tools_util import (
    convert_to_json_schema,
    extract_tool_result_as_sting,
    get_tool_params,
    handle_tools,
    build_tool_response_message,
)

from .pydantic_util import ensure_pydantic_model, build_and_validate

from .known_models import ANTHROPIC_KNOWN_MODELS

from anthropic import (
    AsyncAnthropic,
    NOT_GIVEN,
    NotGiven,
    AnthropicError,
    APIConnectionError,
    APIStatusError,
)
from anthropic.types import MessageParam, ToolParam as AnthropicToolParam
from anthropic.types.message import Message as AnthropicMessage
from anthropic.types.message_create_params import ToolChoice
from anthropic.lib.streaming._types import MessageStreamEvent
from anthropic.types.text_block_param import TextBlockParam
from anthropic.types.image_block_param import ImageBlockParam
from anthropic.types.tool_use_block_param import ToolUseBlockParam
from anthropic.types.tool_result_block_param import ToolResultBlockParam

from openai.lib._pydantic import to_strict_json_schema

from typing import (
    List, Callable, AsyncIterator, Any, Type,
    Tuple, Dict, Union, Literal,
)

import asyncio
import copy
import json

import logging

_LOG = logging.getLogger(__name__)


async def _build_anthropic_content(
    message: MessageContent,
) -> List[Union[TextBlockParam, ImageBlockParam]]:
    ret: List[Union[TextBlockParam, ImageBlockParam]] = []
    if message['text']:
        ret.append({
            'type': 'text',
            'text': message['text'],
        })
    if 'media' in message:
        for m in message['media']:
            if m['type'] == 'image':
                mimetype, data = await convert_to_image_base64(m['image'])
                ret.append({
                    'type': 'image',
                    'source': {
                        'type': 'base64',
                        'media_type': mimetype,
                        'data': data,
                    },
                })
            else:
                raise ValueError(f"unknown media type: {m['type']}")
    return ret


async def _build_anthropic_tool_use(
    message: MessageToolCall,
) -> List[ToolUseBlockParam]:
    ret: List[ToolUseBlockParam] = []
    for tool in message['tools']:
        if tool['call_type'] == 'function':
            ret.append({
                'type': 'tool_use',
                'id': tool['call_id'],
                'name': tool['function']['name'],
                'input': json.loads(tool['function']['arguments']),
            })
        else:
            raise ValueError(f"unknown tool type: {tool['call_type']}")
    if len(ret) == 0:
        raise ValueError(f"no content")
    return ret


async def _build_anthropic_tool_result(
    message: MessageToolResult,
) -> List[ToolResultBlockParam]:
    ret: List[ToolResultBlockParam] = []
    for tool_result in message['tools']:
        content = extract_tool_result_as_sting(tool_result)
        obj: ToolResultBlockParam = {
            'type': 'tool_result',
            'tool_use_id': tool_result['call_id'],
            'content': [
                {
                    'type': 'text',
                    'text': content,
                },
            ],
        }
        if 'is_error' in tool_result and tool_result['is_error']:
            obj['is_error'] = True
        ret.append(obj)
    if len(ret) == 0:
        raise ValueError(f"no content")
    return ret


def _collapse_anthropic_messages(
    messages: List[MessageParam],
) -> List[MessageParam]:
    ms: List[MessageParam] = []
    for i, m in enumerate(messages):
        if i == 0:
            ms.append(m)
            continue
        prev_m = ms[-1]
        if prev_m['role'] == m['role']:
            assert isinstance(prev_m['content'], list)
            assert isinstance(m['content'], list)
            prev_m['content'].extend(m['content'])
        else:
            ms.append(m)
    return ms


def _collapse_tool_call_messages(
    messages: List[Message],
) -> List[Message]:
    ms: List[Message] = []
    for i, m in enumerate(messages):
        if i == 0:
            ms.append(m)
            continue
        prev_m = ms[-1]
        if prev_m['role'] == 'tool_call' and m['role'] == 'tool_call':
            prev_m['tools'].extend(m['tools'])
        else:
            ms.append(m)
    return ms


async def _convert_to_anthropic_messages(
    messages: List[Message],
) -> Tuple[Union[str, None], List[MessageParam]]:
    system_prompt: Union[str, None] = None
    ret: List[MessageParam] = []
    if len(messages) > 0:
        first_message = messages[0]
        if first_message['role'] == 'system':
            if first_message.get('media'):
                raise ValueError(f"For this model, you may not pass media in the system prompt.")
            system_prompt = first_message['text']
            messages = messages[1:]
        for m in messages:
            if m['role'] == 'system':
                raise ValueError(f"For this model, you can only have a system prompt as the first message!")
            elif m['role'] == 'ai':
                ret.append({
                    'role': 'assistant',
                    'content': await _build_anthropic_content(m),
                })
            elif m['role'] == 'human':
                ret.append({
                    'role': 'user',
                    'content': await _build_anthropic_content(m),
                })
            elif m['role'] == 'tool_call':
                ret.append({
                    'role': 'assistant',
                    'content': await _build_anthropic_tool_use(m),
                })
            elif m['role'] == 'tool_res':
                ret.append({
                    'role': 'user',
                    'content': await _build_anthropic_tool_result(m),
                })
            else:
                raise ValueError(f"Unknown role: {m['role']}")
    ret = [r for r in ret if len(list(r['content'])) > 0]
    ret = _collapse_anthropic_messages(ret)
    return system_prompt, ret


def _build_messages_from_anthropic_payload(
    raw_events: List[Dict[str, Any]],
    message: AnthropicMessage,
) -> List[Message]:
    ms: List[Message] = []
    for c in message.content:
        if c.type == 'text':
            ms.append({
                'role': 'ai',
                'text': c.text,
            })
        elif c.type == 'tool_use':
            ms.append({
                'role': 'tool_call',
                'tools': [{
                    'call_id': c.id,
                    'call_type': 'function',
                    'function': {
                        'name': c.name,
                        'arguments': json.dumps(c.input),
                    },
                }],
            })
        else:
            raise ValueError(f"unknown content type: {c.type}")
    ms = _collapse_tool_call_messages(ms)
    if len(ms) == 0:
        ms.append({
            'role': 'ai',
            'text': None,  # <-- sometimes it generates nothing, so we'll show that
        })
    last_message = ms[-1]
    cost: Cost = {
        'input_tokens': message.usage.input_tokens,
        'output_tokens': message.usage.output_tokens,
        'total_tokens': message.usage.input_tokens + message.usage.output_tokens,
    }
    last_message['cost'] = cost
    last_message['raw'] = {
        'events': raw_events,
        'message': message.to_dict(),
    }
    return ms


def _convert_to_anthropic_tool(tool: Callable) -> AnthropicToolParam:
    description, params = get_tool_params(tool)
    return {
        'name': get_name(tool),
        'description': description,
        'input_schema': convert_to_json_schema(params),
    }


def _convert_to_anthropic_tools(tools: List[Callable]) -> Union[NotGiven, List[AnthropicToolParam]]:
    if len(tools) == 0:
        return NOT_GIVEN
    specs = [_convert_to_anthropic_tool(tool) for tool in tools]
    return specs


def _make_raw_event(event: MessageStreamEvent) -> Dict[str, Any]:
    d = event.to_dict()
    if 'snapshot' in d:
        # This accumulated text is overkill.
        # It is O(n^2) in storage, so we don't want this in our database.
        del d['snapshot']
    return d


async def _process_stream(stream: AsyncIterator[MessageStreamEvent]) -> AsyncIterator[EventPayload]:
    content: Dict[int, Tuple[Literal['text', 'tool_use'], str, List[str]]] = {}
    async for event in stream:
        if event.type == 'content_block_start':
            assert event.index not in content
            if event.content_block.type == 'text':
                content[event.index] = event.content_block.type, '', [event.content_block.text]
                if event.content_block.text:
                    yield 'ai', 'text_event', event.content_block.text
            elif event.content_block.type == 'tool_use':
                assert event.content_block.input == {}
                content[event.index] = event.content_block.type, event.content_block.name, []
                yield 'tool_call', 'text_event', f'{event.content_block.name}('
            else:
                raise ValueError(f"unknown content block type: {event.content_block.type}")
        elif event.type == 'content_block_delta':
            assert event.index in content
            if event.delta.type == 'text_delta':
                content[event.index][2].append(event.delta.text)
                if event.delta.text:
                    yield 'ai', 'text_event', event.delta.text
            elif event.delta.type == 'input_json_delta':
                content[event.index][2].append(event.delta.partial_json)
                if event.delta.partial_json:
                    yield 'tool_call', 'text_event', event.delta.partial_json
            else:
                raise ValueError(f"unknown delta type: {event.delta.type}")
        elif event.type == 'content_block_stop':
            assert event.index in content
            if event.content_block.type == 'text':
                pass  # nothing more to do... we already streamed all the text
            elif event.content_block.type == 'tool_use':
                yield 'tool_call', 'text_event', ')\n'
                tool_call: ToolCall = {
                    'call_type': 'function',
                    'call_id': event.content_block.id,
                    'function': {
                        'name': event.content_block.name,
                        'arguments': json.dumps(event.content_block.input),
                    },
                }
                yield 'tool_call', 'tool_call_event', tool_call
            else:
                raise ValueError(f"unknown content block type: {event.content_block.type}")
        else:
            # The other events we can ignore, I think.
            pass


def _log_dumps(val: Any) -> str:
    if isinstance(val, dict):
        return json.dumps(val)
    else:
        return str(val)


class LasagnaAnthropic(Model):
    def __init__(self, model: str, **model_kwargs: Any):
        known_model_names = [m['formal_name'] for m in ANTHROPIC_KNOWN_MODELS]
        if model not in known_model_names:
            _LOG.warning(f'untested model: {model} (may or may not work)')
        self.model = model
        self.model_kwargs = copy.deepcopy(model_kwargs)
        self.n_retries: int = self.model_kwargs.get('retries', 3)
        if not isinstance(self.n_retries, int) or self.n_retries < 0:
            raise ValueError(f"model_kwargs['retries'] must be a non-negative integer (got {self.n_retries})")
        self.model_spec: ModelSpec = {
            'provider': 'anthropic',
            'model': self.model,
            'model_kwargs': self.model_kwargs,
        }

    def config_hash(self) -> str:
        return recursive_hash(None, self.model_spec)

    def _make_client(self) -> AsyncAnthropic:
        client = AsyncAnthropic(
            api_key  = self.model_kwargs.get('api_key', None),
            base_url = self.model_kwargs.get('base_url', None),
        )
        return client

    async def _run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[NotGiven, List[AnthropicToolParam]],
        force_tool: bool,
        disable_parallel_tool_use: bool,
    ) -> List[Message]:
        tool_choice: Union[NotGiven, ToolChoice]
        if force_tool:
            if not tools_spec or len(tools_spec) == 0:
                raise ValueError(f"When `force_tool` is set, you must pass at least one tool!")
            elif len(tools_spec) == 1:
                tool_choice = {
                    "type": "tool",
                    "name": tools_spec[0]['name'],
                    "disable_parallel_tool_use": disable_parallel_tool_use,
                }
            else:
                tool_choice = {
                    "type": "any",   # <-- model must use a tool, but is allowed to choose which one on its own
                    "disable_parallel_tool_use": disable_parallel_tool_use,
                }
        else:
            if tools_spec and len(tools_spec) > 0:
                tool_choice = {
                    "type": "auto",  # <-- the model can choose to use them or not
                    "disable_parallel_tool_use": disable_parallel_tool_use,
                }
            else:
                tool_choice = NOT_GIVEN

        system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)

        max_tokens: int = self.model_kwargs.get('max_tokens', 4096)
        stop: Union[List[str], NotGiven] = self.model_kwargs.get('stop', NOT_GIVEN)
        temperature: Union[float, NotGiven] = float(self.model_kwargs['temperature']) if 'temperature' in self.model_kwargs else NOT_GIVEN
        top_p: Union[float, NotGiven] = float(self.model_kwargs['top_p']) if 'top_p' in self.model_kwargs else NOT_GIVEN
        top_k: Union[int, NotGiven] = self.model_kwargs.get('top_k', NOT_GIVEN)
        user: Union[str, None] = self.model_kwargs.get('user', None)

        assert isinstance(max_tokens, int)
        if stop is not NOT_GIVEN:
            assert isinstance(stop, list)
            for s in stop:
                assert isinstance(s, str)
        if temperature is not NOT_GIVEN:
            assert isinstance(temperature, float)
        if top_p is not NOT_GIVEN:
            assert isinstance(top_p, float)
        if top_k is not NOT_GIVEN:
            assert isinstance(top_k, int)
        if user is not None:
            assert isinstance(user, str)

        _LOG.info(f"Invoking {self.model} with:\n  system_prompt: {system_prompt}\n  messages: {_log_dumps(anthropic_messages)}\n  tools: {_log_dumps(tools_spec)}\n  tool_choice: {tool_choice}")

        client = self._make_client()
        async with client.messages.stream(
            model       = self.model,
            system      = system_prompt or NOT_GIVEN,
            messages    = anthropic_messages,
            max_tokens  = max_tokens,
            tools       = tools_spec,
            tool_choice = tool_choice,
            temperature = temperature,
            top_p       = top_p,
            top_k       = top_k,
            metadata    = {
                'user_id': user,
            },
            stop_sequences = stop,
            # Anthropic doesn't support sending logprobs 👎
        ) as stream:
            raw_stream, rt_stream = adup(stream)
            async for event in _process_stream(rt_stream):
                await event_callback(event)
            raw_events = [_make_raw_event(v) async for v in raw_stream]
            anthropic_message = await stream.get_final_message()
            new_messages = _build_messages_from_anthropic_payload(raw_events, anthropic_message)

        _LOG.info(f"Finished {self.model} with usage: {_log_dumps(new_messages[-1].get('cost'))}")

        return new_messages

    async def _retrying_run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[NotGiven, List[AnthropicToolParam]],
        force_tool: bool,
        disable_parallel_tool_use: bool,
    ) -> List[Message]:
        last_error: Union[Exception, None] = None
        assert self.n_retries + 1 > 0   # <-- we know this is true from the check in __init__
        for delay_on_error in exponential_backoff_retry_delays(self.n_retries + 1):
            try:
                await event_callback(('transaction', 'start', ('anthropic', self.model)))
                try:
                    new_messages = await self._run_once(
                        event_callback = event_callback,
                        messages = messages,
                        tools_spec = tools_spec,
                        force_tool = force_tool,
                        disable_parallel_tool_use = disable_parallel_tool_use,
                    )
                except:
                    await event_callback(('transaction', 'rollback', None))
                    raise
                await event_callback(('transaction', 'commit', None))
                return new_messages
            except AnthropicError as e:
                # Some errors should be retried, some should not. Below
                # is the logic to decide when to retry vs when to not.
                # It's likely this will change as we get more usage and see
                # where Anthropic tends to fail, and when we know more what is
                # recoverable vs not.
                last_error = e
                if isinstance(e, APIConnectionError):
                    # Network connection error. We can retry this.
                    pass
                elif isinstance(e, APIStatusError):
                    if 400 <= e.status_code < 500:
                        # This is a request error. Not recoverable.
                        raise
                    elif e.status_code >= 500:
                        # Server error, so we can retry this.
                        pass
                    else:
                        # Something else??? This is weird, maybe this is recoverable?
                        _LOG.warning(f"Got an unknown error status: {e}")
                        pass
                else:
                    # Some other error that we don't know about. Let's bail.
                    raise
                if delay_on_error > 0.0:
                    _LOG.warning(f"Got a maybe-recoverable error (will retry in {delay_on_error:.2f} seconds): {e}")
                    await asyncio.sleep(delay_on_error)
        assert last_error is not None   # <-- we know this is true because `n_retries + 1 > 0`
        raise last_error

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        messages = [*messages]  # shallow copy
        new_messages: List[Message] = []
        tools_spec = _convert_to_anthropic_tools(tools)
        tools_map = {get_name(tool): tool for tool in tools}
        for _ in range(max_tool_iters):
            new_messages_here = await self._retrying_run_once(
                event_callback = event_callback,
                messages       = messages,
                tools_spec     = tools_spec,
                force_tool     = force_tool,
                disable_parallel_tool_use = False,
            )
            tools_results = await handle_tools(
                prev_messages = messages,
                new_messages = new_messages_here,
                tools_map = tools_map,
                event_callback = event_callback,
                model_spec = self.model_spec,
            )
            new_messages.extend(new_messages_here)
            messages.extend(new_messages_here)
            if tools_results is None:
                break
            for tool_result in tools_results:
                await event_callback(('tool_res', 'tool_res_event', tool_result))
            tool_response_message = build_tool_response_message(tools_results)
            new_messages.append(tool_response_message)
            messages.append(tool_response_message)
        return new_messages

    async def extract(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        extraction_type: Type[ExtractionType],
    ) -> Tuple[Message, ExtractionType]:
        tools_spec: List[AnthropicToolParam] = [
            {
                'name': get_name(extraction_type),
                'input_schema': to_strict_json_schema(ensure_pydantic_model(extraction_type)),
            },
        ]

        docstr = getattr(extraction_type, '__doc__', None)
        if docstr:
            tools_spec[0]['description'] = docstr

        new_messages = await self._retrying_run_once(
            event_callback = event_callback,
            messages       = messages,
            tools_spec     = tools_spec,
            force_tool     = True,
            disable_parallel_tool_use = True,
        )

        assert len(new_messages) == 1
        new_message = new_messages[0]

        if new_message['role'] == 'tool_call':
            tools = new_message['tools']

            assert len(tools) == 1
            parsed = json.loads(tools[0]['function']['arguments'])
            result = build_and_validate(extraction_type, parsed)

            return new_message, result

        else:
            assert new_message['role'] == 'ai'
            text = new_message['text']
            raise RuntimeError(f"Model failed to generate structured output; instead, it output: {text}")



"""
This module is the Lasagna adapter for AWS Bedrock.

For more information about the AWS Bedrock, see:
 - https://docs.aws.amazon.com/bedrock/
"""

from .types import (
    MessageContent,
    MessageToolCall,
    MessageToolResult,
    ModelSpec,
    Message,
    EventCallback,
    Model,
    ExtractionType,
    Cost,
    ToolCall,
)

from .util import (
    convert_to_image_base64,
    exponential_backoff_retry_delays,
    get_name,
    recursive_hash,
)

from .tools_util import (
    convert_to_json_schema,
    extract_tool_result_as_sting,
    get_tool_params,
    handle_tools,
    build_tool_response_message,
)

from .pydantic_util import build_and_validate, ensure_pydantic_model

from .known_models import BEDROCK_KNOWN_MODELS

from typing import (
    List, Callable, Type, AsyncIterable, Any,
    Tuple, Dict, Union,
)

import botocore.client  # type: ignore
import boto3.session  # type: ignore
from botocore.exceptions import ClientError  # type: ignore

from openai.lib._pydantic import to_strict_json_schema

import os
import asyncio
import copy
import json

import logging

_LOG = logging.getLogger(__name__)


def _convert_to_bedrock_tool(tool: Callable) -> Dict:
    description, params = get_tool_params(tool)
    schema = convert_to_json_schema(params)
    if 'additionalProperties' in schema:
        del schema['additionalProperties']
    return {
        'toolSpec': {
            'name': get_name(tool),
            'description': description,
            'inputSchema': {
                'json': schema,
            },
        },
    }


def _convert_to_bedrock_tools(tools: List[Callable]) -> Union[None, List[Dict]]:
    if len(tools) == 0:
        return None
    specs = [_convert_to_bedrock_tool(tool) for tool in tools]
    return specs


async def _build_bedrock_content(message: MessageContent) -> List[Dict]:
    r: List[Dict] = []
    if message['text']:
        r.append({
            'text': message['text'],
        })
    if 'media' in message:
        for m in message['media']:
            assert m['type'] == 'image', 'we only support images, for now'
            mimetype, data = await convert_to_image_base64(m['image'])
            assert mimetype.startswith('image/')
            mimetype = mimetype.split('/')[1]
            assert mimetype in ['png', 'jpeg', 'gif', 'webp']
            r.append({
                'image': {
                    'format': mimetype,
                    'source': {
                        'bytes': data,
                    },
                },
            })
    return r


async def _build_bedrock_tool_use(message: MessageToolCall) -> List[Dict]:
    ret: List[Dict] = []
    for tool in message['tools']:
        if tool['call_type'] == 'function':
            ret.append({
                'toolUse': {
                    'toolUseId': tool['call_id'],
                    'name': tool['function']['name'],
                    'input': json.loads(tool['function']['arguments']),
                },
            })
        else:
            raise ValueError(f"unknown tool type: {tool['call_type']}")
    if len(ret) == 0:
        raise ValueError(f"no content")
    return ret


async def _build_bedrock_tool_result(message: MessageToolResult) -> List[Dict]:
    ret: List[Dict] = []
    for tool_result in message['tools']:
        content = extract_tool_result_as_sting(tool_result)
        is_error = 'is_error' in tool_result and tool_result['is_error']
        ret.append({
            'toolResult': {
                'toolUseId': tool_result['call_id'],
                'content': [
                    {
                        'text': content,
                    },
                ],
                'status': 'error' if is_error else 'success',
            },
        })
    if len(ret) == 0:
        raise ValueError(f"no content")
    return ret


def _collapse_bedrock_messages(messages: List[Dict]) -> List[Dict]:
    ms: List[Dict] = []
    for i, m in enumerate(messages):
        if i == 0:
            ms.append(m)
            continue
        prev_m = ms[-1]
        if prev_m['role'] == m['role']:
            assert isinstance(prev_m['content'], list)
            assert isinstance(m['content'], list)
            prev_m['content'].extend(m['content'])
        else:
            ms.append(m)
    return ms


async def _convert_to_bedrock_messages(
    messages: List[Message],
) -> Tuple[Union[None, str], List[Dict]]:
    system_prompt: Union[str, None] = None
    ret: List[Dict] = []
    if len(messages) > 0:
        first_message = messages[0]
        if first_message['role'] == 'system':
            if first_message.get('media'):
                raise ValueError(f"For this model, you may not pass media in the system prompt.")
            system_prompt = first_message['text']
            messages = messages[1:]
        for m in messages:
            if m['role'] == 'system':
                raise ValueError(f"For this model, you can only have a system prompt as the first message!")
            elif m['role'] == 'ai':
                ret.append({
                    'role': 'assistant',
                    'content': await _build_bedrock_content(m),
                })
            elif m['role'] == 'human':
                ret.append({
                    'role': 'user',
                    'content': await _build_bedrock_content(m),
                })
            elif m['role'] == 'tool_call':
                ret.append({
                    'role': 'assistant',
                    'content': await _build_bedrock_tool_use(m),
                })
            elif m['role'] == 'tool_res':
                ret.append({
                    'role': 'user',
                    'content': await _build_bedrock_tool_result(m),
                })
            else:
                raise ValueError(f"Unknown role: {m['role']}")
    ret = [r for r in ret if len(list(r['content'])) > 0]
    ret = _collapse_bedrock_messages(ret)
    return system_prompt, ret


def _collapse_tool_call_messages(
    messages: List[Message],
) -> List[Message]:
    ms: List[Message] = []
    for i, m in enumerate(messages):
        if i == 0:
            ms.append(m)
            continue
        prev_m = ms[-1]
        if prev_m['role'] == 'tool_call' and m['role'] == 'tool_call':
            prev_m['tools'].extend(m['tools'])
        else:
            ms.append(m)
    return ms


async def _process_bedrock_stream(
    stream: AsyncIterable[Dict],
    event_callback: EventCallback,
) -> List[Message]:
    raw_events: List[Dict] = []
    partial_messages: List[Dict] = []

    def _init_message(role: str) -> None:
        cost: Cost = {
            'input_tokens': 0,
            'output_tokens': 0,
            'total_tokens': 0,
        }
        partial_messages.append({
            'role': role,
            'blocks': {},
            'cost': cost,
        })

    def _get_block(curr_message: Dict, this_block_index: int) -> Dict:
        if this_block_index not in curr_message['blocks']:
            curr_message['blocks'][this_block_index] = {
                'type': 'text',  # <-- default to text unless proven otherwise later
                'content': [],
            }
        this_block = curr_message['blocks'][this_block_index]
        assert isinstance(this_block, dict)
        return this_block

    async for event in stream:
        raw_events.append(event)

        # Frustratingly, Bedrock sometimes sends a `contentBlockDelta`
        # without first sending `messageStart` nor `contentBlockStart`.
        # So, we have to be ready for that!

        if 'messageStart' in event:
            _init_message(event['messageStart']['role'])

        elif 'messageStop' in event:
            pass   # nothing to do here

        elif 'contentBlockStart' in event:
            payload = event['contentBlockStart']
            if len(partial_messages) == 0:
                _init_message('assistant')
            curr_message = partial_messages[-1]
            this_block = _get_block(curr_message, payload['contentBlockIndex'])

            if 'toolUse' in payload['start']:
                tool_use = payload['start']['toolUse']
                call_id = tool_use['toolUseId']
                func_name = tool_use['name']
                this_block['type'] = 'tool_use'
                text = f'{func_name}('
                await event_callback(('tool_call', 'text_event', text))
                this_block['call_id'] = call_id
                this_block['func_name'] = func_name
            else:
                _LOG.warning(f'unknown bedrock `start` event: {event}')

        elif 'contentBlockDelta' in event:
            payload = event['contentBlockDelta']
            if len(partial_messages) == 0:
                _init_message('assistant')
            curr_message = partial_messages[-1]
            this_block = _get_block(curr_message, payload['contentBlockIndex'])

            delta = payload['delta']
            if 'text' in delta:
                assert this_block['type'] == 'text'
                text = delta['text']
                await event_callback(('ai', 'text_event', text))
                this_block['content'].append(text)
            elif 'toolUse' in delta:
                assert this_block['type'] == 'tool_use'
                text = delta['toolUse']['input']
                await event_callback(('tool_call', 'text_event', text))
                this_block['content'].append(text)
            else:
                _LOG.warning(f'unknown bedrock `delta` type: {event}')

        elif 'contentBlockStop' in event:
            payload = event['contentBlockStop']
            if len(partial_messages) == 0:
                _init_message('assistant')
            curr_message = partial_messages[-1]
            this_block_index = payload['contentBlockIndex']
            assert this_block_index in curr_message['blocks']
            this_block = curr_message['blocks'][this_block_index]

            if this_block['type'] == 'tool_use':
                text = ')\n'
                await event_callback(('tool_call', 'text_event', text))
                tool_call: ToolCall = {
                    'call_type': 'function',
                    'call_id': this_block['call_id'],
                    'function': {
                        'name': this_block['func_name'],
                        'arguments': ''.join(this_block['content']),
                    },
                }
                await event_callback(('tool_call', 'tool_call_event', tool_call))
                this_block['did_emit'] = True

        elif 'metadata' in event:
            payload = event['metadata']
            if len(partial_messages) == 0:
                _init_message('assistant')
            curr_message = partial_messages[-1]

            usage = payload['usage']
            curr_message['cost']['input_tokens'] += usage['inputTokens']
            curr_message['cost']['output_tokens'] += usage['outputTokens']
            curr_message['cost']['total_tokens'] += usage['totalTokens']

        else:
            _LOG.warning(f'unknown bedrock streaming event: {event}')

    sorted_blocks = []
    for pm in partial_messages:
        assert pm['role'] == 'assistant'
        sorted_blocks.extend([
            pm['blocks'][idx]
            for idx in sorted(pm['blocks'].keys())
        ])

    ms: List[Message] = []
    for block in sorted_blocks:
        if block['type'] == 'text':
            ms.append({
                'role': 'ai',
                'text': ''.join(block['content']),
            })
        elif block['type'] == 'tool_use':
            tool_call_here: ToolCall = {
                'call_type': 'function',
                'call_id': block['call_id'],
                'function': {
                    'name': block['func_name'],
                    'arguments': ''.join(block['content']),
                },
            }
            ms.append({
                'role': 'tool_call',
                'tools': [tool_call_here],
            })
            if not block.get('did_emit', False):
                text = ')\n'
                await event_callback(('tool_call', 'text_event', text))
                await event_callback(('tool_call', 'tool_call_event', tool_call_here))
        else:
            raise ValueError(f"unknown content type: {block['type']}")
    ms = _collapse_tool_call_messages(ms)
    if len(ms) == 0:
        ms.append({
            'role': 'ai',
            'text': None,  # <-- sometimes it generates nothing, so we'll show that
        })

    total_cost: Cost = {
        'input_tokens': 0,
        'output_tokens': 0,
        'total_tokens': 0,
    }
    for pm in partial_messages:
        total_cost['input_tokens'] += pm['cost']['input_tokens']
        total_cost['output_tokens'] += pm['cost']['output_tokens']
        total_cost['total_tokens'] += pm['cost']['total_tokens']

    last_message = ms[-1]
    last_message['cost'] = total_cost
    last_message['raw'] = {
        'events': raw_events,
    }

    return ms


def _log_dumps(val: Any) -> str:
    if isinstance(val, dict):
        return json.dumps(val)
    else:
        return str(val)


class LasagnaBedrock(Model):
    def __init__(self, model: str, **model_kwargs: Any):
        known_model_names = [m['formal_name'] for m in BEDROCK_KNOWN_MODELS]
        if model not in known_model_names:
            _LOG.warning(f'untested model: {model} (may or may not work)')
        self.model = model
        self.model_kwargs = copy.deepcopy(model_kwargs)
        self.n_retries: int = self.model_kwargs.get('retries', 3)
        if not isinstance(self.n_retries, int) or self.n_retries < 0:
            raise ValueError(f"model_kwargs['retries'] must be a non-negative integer (got {self.n_retries})")
        self.model_spec: ModelSpec = {
            'provider': 'bedrock',
            'model': self.model,
            'model_kwargs': self.model_kwargs,
        }

        # The boto3 client!
        self.session = boto3.session.Session(
            aws_access_key_id     = self.model_kwargs.get('aws_access_key_id',     os.environ.get('AWS_ACCESS_KEY_ID')),
            aws_secret_access_key = self.model_kwargs.get('aws_secret_access_key', os.environ.get('AWS_SECRET_ACCESS_KEY')),
            aws_session_token     = self.model_kwargs.get('aws_session_token',     os.environ.get('AWS_SESSION_TOKEN')),
            region_name           = self.model_kwargs.get('region_name',           os.environ.get('AWS_REGION')),
            profile_name          = self.model_kwargs.get('profile_name',          os.environ.get('AWS_PROFILE')),
        )
        self.config = botocore.client.Config(**self.model_kwargs.get('aws_client_config', {}))
        self.bedrock_client = self.session.client(service_name='bedrock-runtime', config=self.config)
        # Note: `bedrock_client` is thread-safe, so you *can* safely pass
        #       it around to the event loop's thread pool.
        # See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/clients.html#multithreading-or-multiprocessing-with-clients

    def config_hash(self) -> str:
        return recursive_hash(None, self.model_spec)

    async def _run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[None, List[Dict]],
        force_tool: bool,
    ) -> List[Message]:
        system_prompt, bedrock_messages = await _convert_to_bedrock_messages(messages)

        system = [{'text' : system_prompt}] if system_prompt is not None else None

        tool_config: Union[None, Dict] = None
        if tools_spec:
            tool_config = {
                'tools': tools_spec,
            }
            if force_tool:
                if len(tools_spec) == 1:
                    name = tools_spec[0]['toolSpec']['name']
                    tool_config['toolChoice'] = {'tool' : {'name' : name}}
                else:
                    tool_config['toolChoice'] = {'any': {}}
            else:
                tool_config['toolChoice'] = {'auto' : {}}
        elif force_tool:
            raise ValueError(f"When `force_tool` is set, you must pass at least one tool!")

        max_tokens: int = self.model_kwargs.get('max_tokens', 4096)
        stop: Union[List[str], None] = self.model_kwargs.get('stop', None)
        temperature: Union[float, None] = float(self.model_kwargs['temperature']) if 'temperature' in self.model_kwargs else None
        top_p: Union[float, None] = float(self.model_kwargs['top_p']) if 'top_p' in self.model_kwargs else None

        inference_config: Dict[str, Union[int, float, List[str]]] = {}
        assert isinstance(max_tokens, int)
        inference_config['maxTokens'] = max_tokens
        if stop is not None:
            assert isinstance(stop, list)
            for s in stop:
                assert isinstance(s, str)
            inference_config['stopSequences'] = stop
        if temperature is not None:
            inference_config['temperature'] = temperature
        if top_p is not None:
            inference_config['topP'] = top_p

        _LOG.info(f"Invoking {self.model} with:\n  system_prompt: {system_prompt}\n  messages: {_log_dumps(bedrock_messages)}\n  tools: {_log_dumps(tools_spec)}\n  force_tool: {force_tool}")

        params = {
            'modelId': self.model,
            'messages': bedrock_messages,
            'inferenceConfig': inference_config,
        }

        if system is not None:
            params['system'] = system

        if tool_config is not None:
            params['toolConfig'] = tool_config

        loop = asyncio.get_running_loop()
        queue: asyncio.Queue[Union[Dict, None, Exception]] = asyncio.Queue()

        def _run_bedrock_model() -> None:
            try:
                response = self.bedrock_client.converse_stream(**params)
                stream = response.get('stream')
                assert stream, 'Expected to get a stream!'
                for event in stream:
                    assert isinstance(event, dict)
                    loop.call_soon_threadsafe(queue.put_nowait, event)
                loop.call_soon_threadsafe(queue.put_nowait, None)
            except Exception as e:
                loop.call_soon_threadsafe(queue.put_nowait, e)

        task = loop.run_in_executor(None, _run_bedrock_model)

        async def _generator() -> AsyncIterable[Dict]:
            while True:
                event = await queue.get()
                if event is None:
                    break
                elif isinstance(event, Exception):
                    raise event
                yield event

        new_messages = await _process_bedrock_stream(_generator(), event_callback)

        await task

        _LOG.info(f"Finished {self.model} with usage: {_log_dumps(new_messages[-1].get('cost'))}")

        return new_messages

    async def _retrying_run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools_spec: Union[None, List[Dict]],
        force_tool: bool,
    ) -> List[Message]:
        last_error: Union[ClientError, None] = None
        assert self.n_retries + 1 > 0   # <-- we know this is true from the check in __init__
        for delay_on_error in exponential_backoff_retry_delays(self.n_retries + 1):
            try:
                await event_callback(('transaction', 'start', ('bedrock', self.model)))
                try:
                    new_messages = await self._run_once(
                        event_callback = event_callback,
                        messages = messages,
                        tools_spec = tools_spec,
                        force_tool = force_tool,
                    )
                except:
                    await event_callback(('transaction', 'rollback', None))
                    raise
                await event_callback(('transaction', 'commit', None))
                return new_messages
            except ClientError as e:
                # Some errors should be retried, some should not. Below
                # is the logic to decide when to retry vs when to not.
                # It's likely this will change as we get more usage and see
                # where Bedrock tends to fail, and when we know more what is
                # recoverable vs not.
                last_error = e
                status = e.response['ResponseMetadata']['HTTPStatusCode']
                if status in [403, 404, 400]:
                    raise   # not recoverable!
                if delay_on_error > 0.0:
                    _LOG.warning(f"Got a maybe-recoverable error (will retry in {delay_on_error:.2f} seconds): {e}")
                    await asyncio.sleep(delay_on_error)
        assert last_error is not None   # <-- we know this is true because `n_retries + 1 > 0`
        raise last_error

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        messages = [*messages]  # shallow copy
        new_messages: List[Message] = []
        tools_spec = _convert_to_bedrock_tools(tools)
        tools_map = {get_name(tool): tool for tool in tools}
        for _ in range(max_tool_iters):
            new_messages_here = await self._retrying_run_once(
                event_callback = event_callback,
                messages       = messages,
                tools_spec     = tools_spec,
                force_tool     = force_tool,
            )
            tools_results = await handle_tools(
                prev_messages = messages,
                new_messages = new_messages_here,
                tools_map = tools_map,
                event_callback = event_callback,
                model_spec = self.model_spec,
            )
            new_messages.extend(new_messages_here)
            messages.extend(new_messages_here)
            if tools_results is None:
                break
            for tool_result in tools_results:
                await event_callback(('tool_res', 'tool_res_event', tool_result))
            tool_response_message = build_tool_response_message(tools_results)
            new_messages.append(tool_response_message)
            messages.append(tool_response_message)
        return new_messages

    async def extract(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        extraction_type: Type[ExtractionType],
    ) -> Tuple[Message, ExtractionType]:
        tools_spec: List[Dict] = [
            {
                'toolSpec': {
                    'name': get_name(extraction_type),
                    #'description': ... filled below ...
                    'inputSchema': {
                        'json': to_strict_json_schema(ensure_pydantic_model(extraction_type)),
                    },
                },
            },
        ]

        docstr = getattr(extraction_type, '__doc__', None)
        if docstr:
            tools_spec[0]['toolSpec']['description'] = docstr

        new_messages = await self._retrying_run_once(
            event_callback = event_callback,
            messages       = messages,
            tools_spec     = tools_spec,
            force_tool     = True,
        )

        assert len(new_messages) == 1
        new_message = new_messages[0]

        if new_message['role'] == 'tool_call':
            tools = new_message['tools']

            assert len(tools) == 1
            parsed = json.loads(tools[0]['function']['arguments'])
            result = build_and_validate(extraction_type, parsed)

            return new_message, result

        else:
            assert new_message['role'] == 'ai'
            text = new_message['text']
            raise RuntimeError(f"Model failed to generate structured output; instead, it output: {text}")



from lasagna import (
    known_models,
    build_simple_agent,
)

from lasagna.tui import (
    tui_input_loop,
)

from typing import List, Callable

import asyncio

from dotenv import load_dotenv; load_dotenv()


MODEL_BINDER = known_models.BIND_OPENAI_gpt_4o_mini()


async def main() -> None:
    system_prompt = "You are grumpy."
    tools: List[Callable] = []
    my_agent = build_simple_agent(name = 'agent', tools = tools)
    my_bound_agent = MODEL_BINDER(my_agent)
    await tui_input_loop(my_bound_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())



from typing import List

from .types import ModelRecord

from .agent_util import partial_bind_model


OPENAI_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'gpt-4o-mini-2024-07-18',
        'display_name': 'GPT-4o mini',
    },
    {
        'formal_name': 'gpt-4o-2024-08-06',
        'display_name': 'GPT-4o',
    },
    {
        'formal_name': 'gpt-4o-2024-05-13',
        'display_name': 'GPT-4o',
        'outdated': True,
    },
    {
        'formal_name': 'gpt-4-turbo-2024-04-09',
        'display_name': 'GPT-4',
        'outdated': True,
    },
    {
        'formal_name': 'gpt-3.5-turbo-0125',
        'display_name': 'GPT-3.5',
        'outdated': True,
    },
]

BIND_OPENAI_gpt_4o_mini  = partial_bind_model('openai', 'gpt-4o-mini-2024-07-18')
BIND_OPENAI_gpt_4o       = partial_bind_model('openai', 'gpt-4o-2024-08-06')
BIND_OPENAI_gpt_4_turbo  = partial_bind_model('openai', 'gpt-4-turbo-2024-04-09')
BIND_OPENAI_gpt_35_turbo = partial_bind_model('openai', 'gpt-3.5-turbo-0125')


ANTHROPIC_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'claude-3-5-sonnet-20240620',
        'display_name': 'Claude3.5 Sonnet',
    },
    {
        'formal_name': 'claude-3-opus-20240229',
        'display_name': 'Claude3 Opus',
    },
    {
        'formal_name': 'claude-3-sonnet-20240229',
        'display_name': 'Claude3 Sonnet',
        'outdated': True,
    },
    {
        'formal_name': 'claude-3-haiku-20240307',
        'display_name': 'Claude3 Haiku',
    },
]

BIND_ANTHROPIC_claude_35_sonnet = partial_bind_model('anthropic', 'claude-3-5-sonnet-20240620')
BIND_ANTHROPIC_claude_3_opus    = partial_bind_model('anthropic', 'claude-3-opus-20240229')
BIND_ANTHROPIC_claude_3_sonnet  = partial_bind_model('anthropic', 'claude-3-sonnet-20240229')
BIND_ANTHROPIC_claude_3_haiku   = partial_bind_model('anthropic', 'claude-3-haiku-20240307')


OLLAMA_KNOWN_MODELS: List[ModelRecord] = [
    # We'll only list models here that (1) support tool-calling because
    # that's kind of the whole point of lasagna, and (2) we've tested
    # ourselves and pass our "vibe" test. Users are of course more than
    # welcome to use *other* Ollama models as they see fit.
    {
        'formal_name': 'llama3.2',
        'display_name': 'Meta Llama 3.2',
    },
    {
        'formal_name': 'mistral-small',
        'display_name': 'Mistral Small',
    },
    {
        'formal_name': 'mistral-large',
        'display_name': 'Mistral Large',
    },
]

BIND_OLLAMA_llama3_2 = partial_bind_model('ollama', 'llama3.2')
BIND_OLLAMA_mistral_small = partial_bind_model('ollama', 'mistral-small')
BIND_OLLAMA_mistral_large = partial_bind_model('ollama', 'mistral-large')


BEDROCK_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'us.anthropic.claude-3-haiku-20240307-v1:0',
        'display_name': 'Claude 3 Haiku',
    },
    {
        'formal_name': 'us.anthropic.claude-3-sonnet-20240229-v1:0',
        'display_name': 'Claude 3 Sonnet',
    },
    {
        'formal_name': 'us.anthropic.claude-3-opus-20240229-v1:0',
        'display_name': 'Claude 3 Opus',
    },
    {
        'formal_name': 'us.anthropic.claude-3-5-haiku-20241022-v1:0',
        'display_name': 'Claude 3.5 Haiku',
    },
    {
        'formal_name': 'us.anthropic.claude-3-5-sonnet-20240620-v1:0',
        'display_name': 'Claude 3.5 Sonnet',
    },
    {
        'formal_name': 'us.anthropic.claude-3-5-sonnet-20241022-v2:0',
        'display_name': 'Claude 3.5 Sonnet v2',
    },
    {
        'formal_name': 'us.anthropic.claude-3-7-sonnet-20250219-v1:0',
        'display_name': 'Claude 3.7 Sonnet',
    },
]

BIND_BEDROCK_Claude_3_Haiku       = partial_bind_model('bedrock', 'us.anthropic.claude-3-haiku-20240307-v1:0')
BIND_BEDROCK_Claude_3_Sonnet      = partial_bind_model('bedrock', 'us.anthropic.claude-3-sonnet-20240229-v1:0')
BIND_BEDROCK_Claude_3_Opus        = partial_bind_model('bedrock', 'us.anthropic.claude-3-opus-20240229-v1:0')
BIND_BEDROCK_Claude_3_5_Haiku     = partial_bind_model('bedrock', 'us.anthropic.claude-3-5-haiku-20241022-v1:0')
BIND_BEDROCK_Claude_3_5_Sonnet    = partial_bind_model('bedrock', 'us.anthropic.claude-3-5-sonnet-20240620-v1:0')
BIND_BEDROCK_Claude_3_5_Sonnet_v2 = partial_bind_model('bedrock', 'us.anthropic.claude-3-5-sonnet-20241022-v2:0')
BIND_BEDROCK_Claude_3_7_Sonnet    = partial_bind_model('bedrock', 'us.anthropic.claude-3-7-sonnet-20250219-v1:0')


NVIDIA_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'meta/llama3-70b-instruct',
        'display_name': 'meta/llama3-70b-instruct',
    },
    {
        'formal_name': 'meta/llama3-8b-instruct',
        'display_name': 'meta/llama3-8b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.1-8b-instruct',
        'display_name': 'meta/llama-3.1-8b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.1-70b-instruct',
        'display_name': 'meta/llama-3.1-70b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.1-405b-instruct',
        'display_name': 'meta/llama-3.1-405b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-1b-instruct',
        'display_name': 'meta/llama-3.2-1b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-3b-instruct',
        'display_name': 'meta/llama-3.2-3b-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-11b-vision-instruct',
        'display_name': 'meta/llama-3.2-11b-vision-instruct',
    },
    {
        'formal_name': 'meta/llama-3.2-90b-vision-instruct',
        'display_name': 'meta/llama-3.2-90b-vision-instruct',
    },
    {
        'formal_name': 'mistralai/mistral-large',
        'display_name': 'mistralai/mistral-large',
    },
    {
        'formal_name': 'mistralai/codestral-22b-instruct-v0.1',
        'display_name': 'mistralai/codestral-22b-instruct-v0.1',
    },
    {
        'formal_name': 'mistralai/mixtral-8x22b-instruct-v0.1',
        'display_name': 'mistralai/mixtral-8x22b-instruct-v0.1',
    },
    {
        'formal_name': 'mistralai/mixtral-8x7b-instruct-v0.1',
        'display_name': 'mistralai/mixtral-8x7b-instruct-v0.1',
    },
    {
        'formal_name': 'google/gemma-7b',
        'display_name': 'google/gemma-7b',
    },
    {
        'formal_name': 'google/recurrentgemma-2b',
        'display_name': 'google/recurrentgemma-2b',
    },
    {
        'formal_name': 'microsoft/phi-3-mini-128k-instruct',
        'display_name': 'microsoft/phi-3-mini-128k-instruct',
    },
]

BIND_NVIDIA_meta_llama3_70b_instruct           = partial_bind_model('nvidia', 'meta/llama3-70b-instruct')
BIND_NVIDIA_meta_llama3_8b_instruct            = partial_bind_model('nvidia', 'meta/llama3-8b-instruct')
BIND_NVIDIA_meta_llama3_1_8b_instruct          = partial_bind_model('nvidia', 'meta/llama-3.1-8b-instruct')
BIND_NVIDIA_meta_llama3_1_70b_instruct         = partial_bind_model('nvidia', 'meta/llama-3.1-70b-instruct')
BIND_NVIDIA_meta_llama3_1_405b_instruct        = partial_bind_model('nvidia', 'meta/llama-3.1-405b-instruct')
BIND_NVIDIA_meta_llama3_2_1b_instruct          = partial_bind_model('nvidia', 'meta/llama-3.2-1b-instruct')
BIND_NVIDIA_meta_llama3_2_3b_instruct          = partial_bind_model('nvidia', 'meta/llama-3.2-3b-instruct')
BIND_NVIDIA_meta_llama3_2_11b_vision_instruct  = partial_bind_model('nvidia', 'meta/llama-3.2-11b-vision-instruct')
BIND_NVIDIA_meta_llama3_2_90b_vision_instruct  = partial_bind_model('nvidia', 'meta/llama-3.2-90b-vision-instruct')

BIND_NVIDIA_mistralai_mistral_large            = partial_bind_model('nvidia', 'mistralai/mistral-large')
BIND_NVIDIA_mistralai_codestral_22b_instruct   = partial_bind_model('nvidia', 'mistralai/codestral-22b-instruct-v0.1')
BIND_NVIDIA_mistralai_mixtral_8x22b_instruct   = partial_bind_model('nvidia', 'mistralai/mixtral-8x22b-instruct-v0.1')
BIND_NVIDIA_mistralai_mixtral_8x7b_instruct    = partial_bind_model('nvidia', 'mistralai/mixtral-8x7b-instruct-v0.1')

BIND_NVIDIA_google_gemma_7b                    = partial_bind_model('nvidia', 'google/gemma-7b')
BIND_NVIDIA_google_recurrentgemma_2b           = partial_bind_model('nvidia', 'google/recurrentgemma-2b')

BIND_NVIDIA_microsoft_phi_3_mini_128k_instruct = partial_bind_model('nvidia', 'microsoft/phi-3-mini-128k-instruct')



from lasagna import (
    known_models,
    build_simple_agent,
    build_standard_message_extractor,
)

from lasagna.tui import (
    tui_input_loop,
)

import asyncio

from dotenv import load_dotenv; load_dotenv()

import sympy as sp    # type: ignore


def evaluate_math_expression(expression: str) -> float:
    """
    This tool evaluates a math expression and returns the result.
    Pass math expression as a string, for example:
     - "3 * 6 + 1"
     - "cos(2 * pi / 3) + log(8)"
     - "(4.5/2) + (6.3/1.2)"
     - ... etc
    :param: expression: str: the math expression to evaluate
    """
    expr = sp.sympify(expression)
    result = float(expr.evalf())
    return result


async def main() -> None:
    math_agent = build_simple_agent(
        name = 'math_agent',
        tools = [
            evaluate_math_expression,
        ],
        message_extractor = build_standard_message_extractor(
            strip_tool_messages = True,
            system_prompt_override = "You are a math assistant.",
        ),
        doc = "Use this tool if the user asks a math question.",
    )
    health_agent = known_models.BIND_ANTHROPIC_claude_35_sonnet()(
        build_simple_agent(
            name = 'health_agent',
            tools = [],
            message_extractor = build_standard_message_extractor(
                strip_tool_messages = True,
                system_prompt_override = "You are a health coach who motivates through fear.",
            ),
            doc = "Use this tool if the user asks a health question.",
        ),
    )
    my_bound_agent = known_models.BIND_OPENAI_gpt_4o_mini()(
        build_simple_agent(
            name = 'root_agent',
            tools = [
                math_agent,
                health_agent,
            ],
        ),
    )
    await tui_input_loop(
        my_bound_agent,
        system_prompt = "You interact with the user and delegate to your tools as needed.",
    )


if __name__ == '__main__':
    asyncio.run(main())



from .agent_runner import run

from .util import get_name

from .types import (
    AgentSpec,
    AgentRun,
    ModelFactory,
    ModelRecord,
    AgentCallable,
    BoundAgentCallable,
    EventCallback,
    EventPayload,
    Message,
    Model,
    ExtractionType,
    ToolResult,
)

from typing import (
    Union, List, Any, Type,
    Callable, Protocol,
)


def bind_model(
    provider: Union[str, ModelFactory],
    model: Union[str, ModelRecord],
    **model_kwargs: Any,
) -> Callable[[AgentCallable], BoundAgentCallable]:
    class ModelBinder():
        def __call__(self, agent: AgentCallable) -> BoundAgentCallable:
            spec: AgentSpec = {
                'agent': agent,
                'provider': provider,
                'model': model,
                'model_kwargs': model_kwargs,
            }
            async def bound_agent(event_callback: EventCallback, prev_runs: List[AgentRun]) -> AgentRun:
                return await run(spec, event_callback, prev_runs)
            for attr in ['__module__', '__qualname__', '__doc__']:
                if hasattr(agent, attr):
                    setattr(bound_agent, attr, getattr(agent, attr))
            bound_agent.__name__ = get_name(agent)
            return bound_agent

        def __str__(self) -> str:
            info = (provider, model, model_kwargs)
            return f'model binder: {info}'

    return ModelBinder()


class PartiallyBoundAgentCallable(Protocol):
    def __call__(
        self,
        **model_kwargs: Any,
    ) -> Callable[[AgentCallable], BoundAgentCallable]: ...


def partial_bind_model(
    provider: Union[str, ModelFactory],
    model: Union[str, ModelRecord],
) -> PartiallyBoundAgentCallable:
    class PartialModelBinder():
        def __call__(
            self,
            **model_kwargs: Any,
        ) -> Callable[[AgentCallable], BoundAgentCallable]:
            return bind_model(provider, model, **model_kwargs)

        def __str__(self) -> str:
            info = (provider, model)
            return f'partial model binder: {info}'

    return PartialModelBinder()


def _extract_messages_from_tool_result(
    tool_result: ToolResult,
) -> List[Message]:
    if tool_result['type'] == 'layered_agent':
        run = tool_result['run']
        return recursive_extract_messages([run], from_layered_agents=True)
    return []


def _recursive_extract_messages_from_tool_res(
    messages: List[Message],
) -> List[Message]:
    ms: List[Message] = []
    for m in messages:
        ms.append(m)
        if m['role'] == 'tool_res':
            for t in m['tools']:
                ms.extend(_extract_messages_from_tool_result(t))
    return ms


def recursive_extract_messages(
    agent_runs: List[AgentRun],
    from_layered_agents: bool,
) -> List[Message]:
    """DFS retrieve all messages within a list of `AgentRuns`."""
    messages: List[Message] = []
    for run in agent_runs:
        if run['type'] == 'messages':
            messages.extend(
                (
                    _recursive_extract_messages_from_tool_res(run['messages'])
                    if from_layered_agents else
                    run['messages']
                ),
            )
        elif run['type'] == 'chain' or run['type'] == 'parallel':
            messages.extend(
                recursive_extract_messages(run['runs'], from_layered_agents=from_layered_agents),
            )
        elif run['type'] == 'extraction':
            messages.extend(
                (
                    _recursive_extract_messages_from_tool_res(run['messages'])
                    if from_layered_agents else
                    run['messages']
                ),
            )
        else:
            raise RuntimeError('unreachable')
    return messages


def extract_last_message(
    agent_run_or_runs: Union[AgentRun, List[AgentRun]],
    from_layered_agents: bool,
) -> Message:
    if isinstance(agent_run_or_runs, list):
        messages = recursive_extract_messages(agent_run_or_runs, from_layered_agents=from_layered_agents)
    else:
        messages = recursive_extract_messages([agent_run_or_runs], from_layered_agents=from_layered_agents)
    if len(messages) == 0:
        raise ValueError('no messages found')
    return messages[-1]


def flat_messages(
    agent_name: str,
    messages: List[Message],
) -> AgentRun:
    return {
        'agent': agent_name,
        'type': 'messages',
        'messages': messages,
    }


def override_system_prompt(
    messages: List[Message],
    system_prompt: str,
) -> List[Message]:
    sp: Message = {
        'role': 'system',
        'text': system_prompt,
    }
    if messages and messages[0]['role'] == 'system':
        return [sp, *messages[1:]]
    else:
        return [sp, *messages]


def strip_tool_calls_and_results(
    messages: List[Message],
) -> List[Message]:
    return [
        m
        for m in messages
        if m['role'] != 'tool_call' and m['role'] != 'tool_res'
    ]


def strip_all_but_last_human_message(
    messages: List[Message],
) -> List[Message]:
    for m in reversed(messages):
        if m['role'] == 'human':
            return [m]
    return []


async def noop_callback(event: EventPayload) -> None:
    # "noop" mean "no operation" means DON'T DO ANYTHING!
    assert event


MessageExtractor = Callable[[List[AgentRun]], List[Message]]


def build_standard_message_extractor(
    extract_from_layered_agents: bool = False,
    keep_only_last_human_message: bool = False,
    strip_tool_messages: bool = False,
    system_prompt_override: Union[str, None] = None,
) -> MessageExtractor:
    def extractor(prev_runs: List[AgentRun]) -> List[Message]:
        messages = recursive_extract_messages(
            agent_runs = prev_runs,
            from_layered_agents = extract_from_layered_agents,
        )
        if keep_only_last_human_message:
            messages = strip_all_but_last_human_message(messages)
        if strip_tool_messages:
            messages = strip_tool_calls_and_results(messages)
        if system_prompt_override:
            messages = override_system_prompt(messages, system_prompt_override)
        return messages

    return extractor


default_message_extractor = build_standard_message_extractor()


def build_simple_agent(
    name: str,
    tools: List[Callable] = [],
    force_tool: bool = False,
    max_tool_iters: int = 5,
    message_extractor: MessageExtractor = default_message_extractor,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class SimpleAgent():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            messages = message_extractor(prev_runs)
            new_messages = await model.run(
                event_callback = event_callback,
                messages = messages,
                tools = tools,
                force_tool = force_tool,
                max_tool_iters = max_tool_iters,
            )
            return flat_messages(name, new_messages)

        def __str__(self) -> str:
            return name

    a = SimpleAgent()
    if doc:
        a.__doc__ = doc
    return a


def build_extraction_agent(
    name: str,
    extraction_type: Type[ExtractionType],
    message_extractor: MessageExtractor = default_message_extractor,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class ExtractionAgent():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            messages = message_extractor(prev_runs)
            message, result = await model.extract(event_callback, messages, extraction_type)
            return {
                'agent': name,
                'type': 'extraction',
                'messages': [message],
                'result': result,
            }

        def __str__(self) -> str:
            return name

    a = ExtractionAgent()
    if doc:
        a.__doc__ = doc
    return a


def build_agent_chainer(
    name: str,
    agents: List[BoundAgentCallable],
    message_extractor: Union[MessageExtractor, None] = None,
    doc: Union[str, None] = None,
) -> BoundAgentCallable:
    class ChainedAgents():
        async def __call__(
            self,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            if message_extractor is not None:
                prev_runs = [
                    flat_messages(
                        agent_name = name,
                        messages = message_extractor(prev_runs),
                    ),
                ]
            else:
                prev_runs = [*prev_runs]  # shallow copy
            new_runs: List[AgentRun] = []
            for agent in agents:
                this_run = await agent(event_callback, prev_runs)
                prev_runs.append(this_run)
                new_runs.append(this_run)
            return {
                'agent': name,
                'type': 'chain',
                'runs': new_runs,
            }

        def __str__(self) -> str:
            return name

    a = ChainedAgents()
    if doc:
        a.__doc__ = doc
    return a


def build_agent_router(
    name: str,
    extraction_type: Type[ExtractionType],
    pick_agent_func: Callable[[ExtractionType], BoundAgentCallable],
    message_extractor: MessageExtractor = default_message_extractor,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class AgentRouter():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            messages = message_extractor(prev_runs)
            message, result = await model.extract(event_callback, messages, extraction_type)
            extraction: AgentRun = {
                'agent': name,
                'type': 'extraction',
                'messages': [message],
                'result': result,
            }
            agent = pick_agent_func(result)
            run = await agent(event_callback, prev_runs)
            return {
                'agent': name,
                'type': 'chain',
                'runs': [
                    extraction,
                    run,
                ],
            }

        def __str__(self) -> str:
            return name

    a = AgentRouter()
    if doc:
        a.__doc__ = doc
    return a


def build_static_output_agent(
    name: str,
    output: str,
    doc: Union[str, None] = None,
) -> AgentCallable:
    class StaticOutputAgent():
        async def __call__(
            self,
            model: Model,
            event_callback: EventCallback,
            prev_runs: List[AgentRun],
        ) -> AgentRun:
            assert model is not None and prev_runs is not None
            await event_callback(('ai', 'text_event', output))
            return flat_messages(name, [
                {
                    'role': 'ai',
                    'text': output,
                },
            ])

        def __str__(self) -> str:
            return name

    a = StaticOutputAgent()
    if doc:
        a.__doc__ = doc
    return a



import asyncio
import inspect
import json
import collections.abc
from enum import Enum

from typing import (
    List, Dict, Tuple, Union, Any, Callable, Literal,
    get_origin, get_args, cast,
)
from typing_extensions import is_typeddict

from .agent_util import bind_model, extract_last_message, flat_messages

from .util import (
    parse_docstring,
    DOCSTRING_PARAM_SUPPORTED_TYPES,
    get_name,
)

from .types import (
    AgentCallable,
    AgentRun,
    BoundAgentCallable,
    EventCallback,
    Message,
    ModelSpec,
    ToolCall,
    ToolResult,
    ToolParam,
)

import logging

_LOG = logging.getLogger(__name__)


def convert_to_json_schema(params: List[ToolParam]) -> Dict[str, object]:
    def convert_type(t: str) -> Dict[str, object]:
        if t.startswith('enum '):
            return {
                "type": "string",
                "enum": t.split()[1:],
            }
        else:
            return {
                "type": {
                    'str': 'string',
                    'float': 'number',
                    'int': 'integer',
                    'bool': 'boolean',
                }[t],
            }
    return {
        "type": "object",
        "properties": {
            p['name']: {
                **convert_type(p['type']),
                **({
                    "description": p['description'],
                } if 'description' in p and p['description'] else {}),
            }
            for p in params
        },
        "required": [
            p['name']
            for p in params
            if not p.get('optional', False)
        ],
        "additionalProperties": False,
    }


def is_async_callable(f: Any) -> bool:
    if not inspect.isfunction(f) and hasattr(f, '__call__'):
        f = f.__call__
    return callable(f) and inspect.iscoroutinefunction(f)


def _grab_correct_docstring(f: Any) -> str:
    outer_doc = getattr(f, '__doc__', None)
    inner_doc = None
    if not inspect.isfunction(f) and hasattr(f, '__call__'):
        inner_doc = getattr(f.__call__, '__doc__', None)
    return inner_doc or outer_doc or ''


def is_callable_of_type(
    concrete_callable: Any,
    expected_callable_type: Any,
    no_throw: bool = False,
) -> bool:
    if get_origin(expected_callable_type) is not collections.abc.Callable:
        if no_throw:
            return False
        raise ValueError("`expected_callable_type` must be a Callable type")

    expected_params, expected_return = get_args(expected_callable_type)

    if not callable(concrete_callable):
        if no_throw:
            return False
        raise ValueError("`concrete_callable` must be a callable object")

    concrete_sig = inspect.signature(concrete_callable)
    concrete_params = concrete_sig.parameters.values()

    for concrete_param in concrete_params:
        if concrete_param.annotation is inspect.Parameter.empty:
            if no_throw:
                return False
            raise ValueError(f"`concrete_callable` must have fully-specified types: missing type for `{concrete_param.name}`")

    if concrete_sig.return_annotation is inspect.Parameter.empty:
        if no_throw:
            return False
        raise ValueError(f"`concrete_callable` must have fully-specified types: missing return type")

    if len(concrete_params) != len(expected_params):
        return False

    for concrete_param, expected_param in zip(concrete_params, expected_params):
        if not type_a_isa_b(expected_param, concrete_param.annotation):
            return False

    if is_async_callable(concrete_callable):
        if get_origin(expected_return) is not collections.abc.Awaitable:
            return False
        eret = get_args(expected_return)[0]
        if not type_a_isa_b(concrete_sig.return_annotation, eret):
            return False

    else:
        if not type_a_isa_b(concrete_sig.return_annotation, expected_return):
            return False

    return True


def type_a_isa_b(a: Any, b: Any) -> bool:
    """
    Correctly compares annotation types such that, e.g., these are equal:
        a = List[dict[str, Set[int]]]
        b = list[Dict[str, set[int]]]

    Also correctly identifies `a isa b`, e.g.
        a = List[str]
        b = Iterable[str]

    Also correctly handles `Union` and `Literal` types.
    """
    a_origin = get_origin(a) or a
    b_origin = get_origin(b) or b
    a_args = get_args(a)
    b_args = get_args(b)

    if b_origin is Any:
        return True

    if a_origin is Union:
        return all(type_a_isa_b(option, b) for option in a_args)
    elif b_origin is Union:
        return any(type_a_isa_b(a, option) for option in b_args)

    if a_origin is Literal and b_origin is Literal:
        return len(set(a_args) - set(b_args)) == 0
    elif a_origin is Literal:
        return all(type_a_isa_b(type(option), b) for option in a_args)
    elif b_origin is Literal:
        _LOG.warning(f'Cannot check against target literal: {a=} {b=}')
        return False

    if is_typeddict(a) and b_origin is dict:
        if not b_args or b_args == (str, Any):
            # Special case we want to handle!
            return True

    try:
        if not (a_origin == b_origin or issubclass(a_origin, b_origin)):
            return False
    except TypeError:
        # Note: Calling `issubclass` on two TypedDicts will land here.
        #       Our equality check above should catch the cases we need, for now.
        #       But, we might need to fix this in the future if we ever need to
        #       compare two TypedDicts that are not the *same* but are *compatible*
        #       with one another.
        return False

    if len(b_args) == 0:
        # Since `b` is totally generic, then `a` must merely be a more
        # specific version, which passes our test.
        return True
    if len(a_args) != len(b_args):
        return False
    for a_arg, b_arg in zip(a_args, b_args):
        if not type_a_isa_b(a_arg, b_arg):
            return False

    return True


def get_tool_params(tool: Callable) -> Tuple[str, List[ToolParam]]:
    doc_params: List[ToolParam]
    description, doc_params = parse_docstring(_grab_correct_docstring(tool))

    is_agent_callable = is_callable_of_type(tool, AgentCallable, no_throw=True)
    is_bound_agent_callable = is_callable_of_type(tool, BoundAgentCallable, no_throw=True)

    if is_agent_callable or is_bound_agent_callable:
        pass  # no validation on these

    else:
        concrete_sig = inspect.signature(tool)
        concrete_params = concrete_sig.parameters.values()

        if len(concrete_params) != len(doc_params):
            raise ValueError(f'tool `{get_name(tool)}` has parameter length mismatch: tool has {len(concrete_params)}, docstring has {len(doc_params)}')

        for concrete_param, doc_param in zip(concrete_params, doc_params):
            if concrete_param.name != doc_param['name']:
                raise ValueError(f"tool `{get_name(tool)}` has parameter name mismatch: tool name is `{concrete_param.name}`, docstring name is `{doc_param['name']}`")
            if doc_param.get('optional', False):
                if concrete_param.default is inspect.Parameter.empty:
                    raise ValueError(f"tool `{get_name(tool)}` has an optional parameter without a default value: `{concrete_param.name}`")
            if concrete_param.annotation is inspect.Parameter.empty:
                _LOG.warning(f'tool `{get_name(tool)}` is missing annotation for parameter `{concrete_param.name}`')
            else:
                if doc_param['type'].startswith('enum '):
                    if issubclass(concrete_param.annotation, str):
                        pass  # a str can accept any value, so we're good
                    elif issubclass(concrete_param.annotation, Enum):
                        concrete_enum_vals = set([
                            e.value
                            for e in concrete_param.annotation.__members__.values()
                        ])
                        doc_enum_vals = set(doc_param['type'].split()[1:])
                        if concrete_enum_vals != doc_enum_vals:
                            raise ValueError(f"tool `{get_name(tool)}` has parameter `{concrete_param.name}` enum value mismatch: tool has enum values `{sorted(concrete_enum_vals)}`, docstring has enum values `{sorted(doc_enum_vals)}`")
                    else:
                        raise ValueError(f"tool `{get_name(tool)}` has parameter `{concrete_param.name}` type mismatch: tool type is `{concrete_param.annotation}`, docstring type is `{doc_param['type']}`")
                else:
                    doc_param_type = DOCSTRING_PARAM_SUPPORTED_TYPES.get(doc_param['type'], 'unknown')
                    if not type_a_isa_b(doc_param_type, concrete_param.annotation):
                        raise ValueError(f"tool `{get_name(tool)}` has parameter `{concrete_param.name}` type mismatch: tool type is `{concrete_param.annotation}`, docstring type is `{doc_param_type}`")

    return description, doc_params


def validate_args(tool: Callable, parsed_args: Dict) -> Dict:
    tool_name = get_name(tool)
    _, doc_params = parse_docstring(_grab_correct_docstring(tool))
    concrete_sig = inspect.signature(tool)
    concrete_params = concrete_sig.parameters.values()

    # Check for unexpected parameters:
    unexpected_params: List[str] = []
    doc_param_names = set([doc_param['name'] for doc_param in doc_params])
    for arg in sorted(parsed_args.keys()):
        if arg not in doc_param_names:
            unexpected_params.append(arg)
    if unexpected_params:
        n = len(unexpected_params)
        s = ', '.join([repr(m) for m in unexpected_params])
        if n == 1:
            raise TypeError(f"{tool_name}() got 1 unexpected argument: {s}")
        else:
            raise TypeError(f"{tool_name}() got {n} unexpected arguments: {s}")

    # Check for missing parameters:
    missing_params: List[str] = []
    for doc_param in doc_params:
        name = doc_param['name']
        optional = doc_param.get('optional', False)
        if not optional and name not in parsed_args:
            missing_params.append(name)
    if missing_params:
        n = len(missing_params)
        s = ', '.join([repr(m) for m in missing_params])
        if n == 1:
            raise TypeError(f"{tool_name}() missing 1 required argument: {s}")
        else:
            raise TypeError(f"{tool_name}() missing {n} required arguments: {s}")

    # Check types:
    validated_args: Dict[str, Any] = {}
    doc_param_type_lookup = {
        doc_param['name']: doc_param['type']
        for doc_param in doc_params
    }
    concrete_annotation_lookup = {
        concrete_param.name: concrete_param.annotation
        for concrete_param in concrete_params
    }
    for arg, value in parsed_args.items():
        assert arg in doc_param_type_lookup
        type_str = doc_param_type_lookup[arg]
        if type_str.startswith('enum '):
            value_str = str(value)
            doc_enum_vals = set(type_str.split()[1:])
            if value_str not in doc_enum_vals:
                raise TypeError(f"{tool_name}() got invalid value for argument `{arg}`: {repr(value)} (valid values are {sorted(doc_enum_vals)})")
            concrete_annotation = concrete_annotation_lookup[arg]
            if concrete_annotation is inspect.Parameter.empty:
                validated_args[arg] = value_str
            elif issubclass(concrete_annotation, str):
                validated_args[arg] = value_str
            else:
                assert issubclass(concrete_annotation, Enum)  # we know this from the tool passing `get_tool_params()`
                concrete_enum_lookup = {
                    e.value: e
                    for e in concrete_annotation.__members__.values()
                }
                assert value_str in concrete_enum_lookup  # we know this from the tool passing `get_tool_params()`
                validated_args[arg] = concrete_enum_lookup[value_str]
        else:
            type_ = DOCSTRING_PARAM_SUPPORTED_TYPES[type_str]
            try:
                validated_args[arg] = type_(value)  # <-- assume c'tor will throw as necessary
            except Exception as e:
                raise TypeError(f"{tool_name}() got invalid value for argument `{arg}`: {repr(value)} ({e})")
    return validated_args


async def _run_single_tool(
    tool_call: ToolCall,
    prev_messages: List[Message],
    tools_map: Dict[str, Callable],
    event_callback: EventCallback,
    model_spec: Union[ModelSpec, None],
) -> ToolResult:
    call_id = 'unknown'

    try:
        call_id = tool_call['call_id']
        func = tools_map[tool_call['function']['name']]
        args = tool_call['function']['arguments']

        parsed_args = json.loads(args)

        if not isinstance(parsed_args, dict):
            raise TypeError("tool output must be a JSON object")

        parsed_args = validate_args(func, parsed_args)

        is_agent_callable = is_callable_of_type(func, AgentCallable, no_throw=True)
        is_bound_agent_callable = is_callable_of_type(func, BoundAgentCallable, no_throw=True)

        if is_agent_callable or is_bound_agent_callable:
            messages: List[Message] = [*prev_messages]  # shallow copy
            if len(parsed_args) > 0:
                messages.append({
                    'role': 'tool_call',
                    'tools': [tool_call],
                })
            prev_runs: List[AgentRun] = [
                flat_messages(
                    agent_name = 'upstream_agent',
                    messages = messages,
                ),
            ]

            if is_agent_callable:
                assert model_spec is not None
                agent = cast(AgentCallable, func)
                binder = bind_model(
                    provider = model_spec['provider'],
                    model = model_spec['model'],
                    **model_spec.get('model_kwargs', {}),
                )
                bound_agent = binder(agent)
            elif is_bound_agent_callable:
                bound_agent = cast(BoundAgentCallable, func)
            else:
                raise RuntimeError('unreachable')

            this_run = await bound_agent(event_callback, prev_runs)

            return {
                'type': 'layered_agent',
                'call_id': call_id,
                'run': this_run,
            }

        else:
            if is_async_callable(func):
                res = await func(**parsed_args)
            else:
                def _wrapped_sync() -> Any:
                    return func(**parsed_args)
                loop = asyncio.get_running_loop()
                res = await loop.run_in_executor(None, _wrapped_sync)
            return {
                'type': 'any',
                'call_id': call_id,
                'result': res,
            }

    except Exception as e:
        _LOG.exception(e)
        error = f"{get_name(type(e))}: {e}"
        return {
            'type': 'any',
            'call_id': call_id,
            'result': error,
            'is_error': True,
        }


async def handle_tools(
    prev_messages: List[Message],
    new_messages: List[Message],
    tools_map: Dict[str, Callable],
    event_callback: EventCallback,
    model_spec: Union[ModelSpec, None],
) -> Union[List[ToolResult], None]:
    tool_messages = [message for message in new_messages if message['role'] == 'tool_call']
    if len(tool_messages) == 0:
        return None

    to_gather: List[asyncio.Task[ToolResult]] = []

    for message in tool_messages:
        for t in message['tools']:
            assert t['call_type'] == 'function'
            to_gather.append(asyncio.create_task(
                _run_single_tool(
                    t,
                    prev_messages,
                    tools_map,
                    event_callback,
                    model_spec,
                ),
            ))

    return await asyncio.gather(*to_gather)


def extract_tool_result_as_sting(tool_result: ToolResult) -> str:
    content: str

    if tool_result['type'] == 'any':
        content = str(tool_result['result'])

    elif tool_result['type'] == 'layered_agent':
        last_message = extract_last_message([tool_result['run']], from_layered_agents=True)
        if last_message['role'] == 'ai':
            content = last_message['text'] or ''
        elif last_message['role'] == 'tool_call':
            content = json.dumps([
                {
                    'name': c['function']['name'],
                    'values': json.loads(c['function']['arguments']),
                }
                for c in last_message['tools']
            ])
        elif last_message['role'] == 'tool_res':
            content = '\n'.join([
                extract_tool_result_as_sting(r)
                for r in last_message['tools']
            ])
        else:
            raise RuntimeError('unreachable')

    else:
        raise RuntimeError('unreachable')

    return content


def build_tool_response_message(tool_results: List[ToolResult]) -> Message:
    return {
        'role': 'tool_res',
        'tools': tool_results,
    }



from lasagna.tui import _truncate_str


def test_truncate_str():
    # String is shorter than allowance:
    t = _truncate_str('hi', 10)
    assert t == 'hi'

    # String equals the allowance:
    t = _truncate_str('hi', 2)
    assert t == 'hi'

    # String is less than allowance, but showing the
    # truncated message would be *longer*:
    t = _truncate_str('hi', 1)
    assert t == 'hi'

    lorem_ipsum = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Morbi sed lacus nisl. Aliquam pharetra lacus non diam porttitor, quis congue libero imperdiet. Vestibulum dapibus est ut urna interdum, eu sagittis metus lobortis. Ut quis tempor dolor, vitae sagittis lectus. Morbi non ex vel leo ullamcorper hendrerit. In eleifend, tortor tincidunt mattis mollis, nibh mauris ultricies velit, convallis fringilla lectus dui non dui. Duis quis elit vitae elit pretium efficitur. Fusce urna magna, interdum quis interdum sit amet, congue a magna. Maecenas vitae aliquam purus. Aliquam erat volutpat. Etiam sit amet pretium massa.'

    # String is shorter than allowance:
    t = _truncate_str(lorem_ipsum, 10000000)
    assert t == lorem_ipsum

    # String equals the allowance:
    t = _truncate_str(lorem_ipsum, len(lorem_ipsum))
    assert t == lorem_ipsum

    # String is less than allowance, but showing the
    # truncated message would be *longer*:
    t = _truncate_str(lorem_ipsum, len(lorem_ipsum) - 5)
    assert t == lorem_ipsum

    # Truncation is necessary:
    t = _truncate_str(lorem_ipsum, 11)
    assert t == 'Lorem ipsum [... truncated 607 characters ...]'



from lasagna import (
    known_models,
    build_simple_agent,
)

from lasagna.tui import (
    tui_input_loop,
)

from typing import List, Callable

import asyncio

from dotenv import load_dotenv; load_dotenv()

import svs


MODEL_BINDER = known_models.BIND_OPENAI_gpt_4o_mini()


class DadJokeTool:
    def __init__(self) -> None:
        self.kb = svs.AsyncKB('https://github.com/Rhobota/svs/raw/main/examples/dad_jokes/dad_jokes.sqlite.gz')

        # It's best to tell the KB to start loading in the background *immediately*.
        # Otherwise it will lazy-load (i.e. wait to start loading until you query it).
        # We'd rather not have our first user wait, so we'll start loading it *now*.
        asyncio.create_task(self.kb.load())

    async def __call__(self, query: str) -> str:
        """
        This tool retrieves "Dad Jokes" from a knowledge base. You pass in a
        search string (`query`) to describe the desired topic, and this tool
        retrieves 5 jokes about that topic. For example, if you want jokes about
        cats, you may pass in the string "cats" as the query.
        :param: query: str: the search string to use to retrieve 5 jokes
        """
        results = await self.kb.retrieve(query, n = 5)
        return '\n\n'.join([
            f"Joke #{i+1}: {result['doc']['text']}"
            for i, result in enumerate(results)
        ])


SYSTEM_PROMPT = ' '.join("""
You are an assistant whose primary job is to tell the user "Dad Jokes", when
asked. A "Dad Joke" is a silly joke that dads are notorious for telling
their kids, much to their kids' dismay (because the jokes are often cheesy
and low-quality). You use the RAG-style tool named `query_dad_jokes` to
retrieve jokes from a curated knowledge base. That tool returns 5 jokes;
when the user asks you for a joke, you'll retrieve those 5 jokes then pick
only the best **one** to give back to the user.
""".strip().splitlines())


async def main() -> None:
    tools: List[Callable] = [
        DadJokeTool(),
    ]
    my_agent = build_simple_agent(name = 'agent', tools = tools)
    my_bound_agent = MODEL_BINDER(my_agent)
    await tui_input_loop(my_bound_agent, SYSTEM_PROMPT)


if __name__ == '__main__':
    asyncio.run(main())



from .types import (
    ExtractionType,
)

from .util import get_name

from typing_extensions import is_typeddict

from typing import (
    Set, Dict, Any, Type,
    get_type_hints, cast,
)

from pydantic import BaseModel, Field, create_model


def create_pydantic_model_from_typeddict(
    typeddict_cls: Type[ExtractionType],
) -> Type[BaseModel]:
    annotations = get_type_hints(typeddict_cls)
    required_keys: Set = getattr(typeddict_cls, '__required_keys__', set())
    fields: Dict[str, Any] = {}
    for field_name, field_type in annotations.items():
        if field_name in required_keys:
            fields[field_name] = (field_type, Field(...))
        else:
            fields[field_name] = (field_type, Field(default=None))
    model_name = get_name(typeddict_cls)
    model: Type[BaseModel] = create_model(model_name, **fields)
    return model


def ensure_pydantic_model(
    extraction_type: Type[ExtractionType],
) -> Type[BaseModel]:
    if issubclass(extraction_type, BaseModel):
        return extraction_type
    elif is_typeddict(extraction_type):
        return create_pydantic_model_from_typeddict(extraction_type)
    else:
        raise ValueError(f'Cannot handle parsing data for type: {extraction_type}')


def build_and_validate(
    extraction_type: Type[ExtractionType],
    data: Any,
) -> ExtractionType:
    if issubclass(extraction_type, BaseModel):
        return cast(ExtractionType, extraction_type.model_validate(data))
    elif is_typeddict(extraction_type):
        model = create_pydantic_model_from_typeddict(extraction_type)
        obj = model.model_validate(data)
        return cast(ExtractionType, obj.model_dump())
    else:
        raise ValueError(f'Cannot handle parsing data for type: {extraction_type}')



import pytest

import os
import tempfile

from typing import List, Dict, Literal
from typing_extensions import TypedDict

from pydantic import BaseModel

from lasagna.types import EventPayload, Message, ToolCall

from lasagna.stream import fake_async

from lasagna.lasagna_ollama import (
    _convert_to_ollama_tool,
    _convert_to_ollama_tools,
    _convert_to_ollama_media,
    _convert_to_ollama_tool_calls,
    _convert_to_ollama_messages,
    _set_cost_raw,
    _process_stream,
    _wrap_event_callback_convert_ai_text_to_tool_call_text,
    _get_ollama_format_for_structured_output,
    LasagnaOllama,
)


def sample_tool(a: int, b: str = 'hi') -> str:
    """
    A sample tool.
    :param: a: int: the value of a
    :param: b: str: (optional) the value of b
    """
    return b * a


_sample_tool_correct_schema = {
    'type': 'function',
    'function': {
        'name': 'sample_tool',
        'description': 'A sample tool.',
        'parameters': {
            'type': 'object',
            'properties': {
                'a': {
                    'type': 'integer',
                    'description': 'the value of a',
                },
                'b': {
                    'type': 'string',
                    'description': '(optional) the value of b',
                },
            },
            'required': ['a'],
            'additionalProperties': False,
        },
    },
}


_sample_events_streaming_text: List[Dict] = [
    {'message': {'role': 'assistant', 'content': 'Hello'}, 'done': False},
    {'message': {'role': 'assistant', 'content': '!'}, 'done': False},
    {'message': {'role': 'assistant', 'content': ' How'}, 'done': False},
    {'message': {'role': 'assistant', 'content': ' can'}, 'done': False},
    {'message': {'role': 'assistant', 'content': ' I'}, 'done': False},
    {'message': {'role': 'assistant', 'content': ' assist'}, 'done': False},
    {'message': {'role': 'assistant', 'content': ' you'}, 'done': False},
    {'message': {'role': 'assistant', 'content': ' today'}, 'done': False},
    {'message': {'role': 'assistant', 'content': '?'}, 'done': False},
    {'message': {'role': 'assistant', 'content': ''}, 'done_reason': 'stop', 'done': True, 'total_duration': 290565553, 'load_duration': 4737247, 'prompt_eval_count': 20, 'prompt_eval_duration': 27000000, 'eval_count': 10, 'eval_duration': 257000000},
]


_sample_events_tool_call: List[Dict] = [
    # Ollama cannot stream in this case, so we get a single message here:
    {'message': {'role': 'assistant', 'content': '', 'tool_calls': [{'function': {'name': 'evaluate_math_expression', 'arguments': {'expression': '2.5 * sin(4.5)'}}}]}, 'done_reason': 'stop', 'done': True, 'total_duration': 799653463, 'load_duration': 4977833, 'prompt_eval_count': 100, 'prompt_eval_duration': 27000000, 'eval_count': 29, 'eval_duration': 765000000},
]


_sample_events_structued_output: List[Dict] = [
    # Ollama cannot stream in this case, so we get a single message here:
    {'message': {'role': 'assistant', 'content': '{"base":2,"exponent":101}'}, 'done_reason': 'stop', 'done': True, 'total_duration': 426267794, 'load_duration': 4418431, 'prompt_eval_count': 28, 'prompt_eval_duration': 21000000, 'eval_count': 15, 'eval_duration': 399000000},
]


def test_convert_to_ollama_tool():
    assert _convert_to_ollama_tool(sample_tool) == _sample_tool_correct_schema


def test_convert_to_ollama_tools():
    assert _convert_to_ollama_tools([]) is None
    assert _convert_to_ollama_tools([sample_tool]) == [
        _sample_tool_correct_schema,
    ]


@pytest.mark.asyncio
async def test_convert_to_ollama_media():
    with tempfile.TemporaryDirectory() as tmp:
        fn1 = os.path.join(tmp, 'a.png')
        fn2 = os.path.join(tmp, 'b.png')
        with open(fn1, 'wb') as f:
            f.write(b'1234')
        with open(fn2, 'wb') as f:
            f.write(b'1235')
        assert await _convert_to_ollama_media([
            {
                'type': 'image',
                'image': fn1,
            },
            {
                'type': 'image',
                'image': fn2,
            },
        ]) == {
            'images': ['MTIzNA==', 'MTIzNQ=='],
        }


def test_convert_to_ollama_tool_calls():
    assert _convert_to_ollama_tool_calls([
        {
            'call_id': 'abc',
            'call_type': 'function',
            'function': {
                'name': 'foo',
                'arguments': '{"a": 7}',
            },
        },
        {
            'call_id': 'xyz',
            'call_type': 'function',
            'function': {
                'name': 'bar',
                'arguments': '{"b": 42}',
            },
        },
    ]) == [
        {
            'function': {
                'name': 'foo',
                'arguments': {
                    'a': 7,
                },
            },
        },
        {
            'function': {
                'name': 'bar',
                'arguments': {
                    'b': 42,
                },
            },
        },
    ]


@pytest.mark.asyncio
async def test_convert_to_ollama_messages():
    with tempfile.TemporaryDirectory() as tmp:
        fn = os.path.join(tmp, 'a.png')
        with open(fn, 'wb') as f:
            f.write(b'1234')
        assert await _convert_to_ollama_messages([
            {
                'role': 'system',
                'text': 'You are a robot.',
            },
            {
                'role': 'human',
                'text': 'What is in this image?',
                'media': [
                    {
                        'type': 'image',
                        'image': fn,
                    },
                ],
            },
            {
                'role': 'ai',
                'text': 'Nothing!',
            },
            {
                'role': 'tool_call',
                'tools': [
                    {
                        'call_id': 'abc',
                        'call_type': 'function',
                        'function': {
                            'name': 'foo',
                            'arguments': '{"a": 7}',
                        },
                    },
                    {
                        'call_id': 'xyz',
                        'call_type': 'function',
                        'function': {
                            'name': 'bar',
                            'arguments': '{"b": 42}',
                        },
                    },
                ],
            },
            {
                'role': 'tool_res',
                'tools': [
                    {
                        'call_id': 'xyz',
                        'type': 'any',
                        'result': True,
                    },
                    {
                        'call_id': 'abc',
                        'type': 'any',
                        'result': 'it worked',
                    },
                    {
                        'call_id': 'dne',
                        'type': 'any',
                        'result': 'ghost result',
                    },
                ],
            },
        ]) == [
            {
                'role': 'system',
                'content': 'You are a robot.',
            },
            {
                'role': 'user',
                'content': 'What is in this image?',
                'images': ['MTIzNA=='],
            },
            {
                'role': 'assistant',
                'content': 'Nothing!',
            },
            {
                'role': 'assistant',
                'content': '',
                'tool_calls': [
                    {
                        'function': {
                            'name': 'foo',
                            'arguments': {
                                'a': 7,
                            },
                        },
                    },
                    {
                        'function': {
                            'name': 'bar',
                            'arguments': {
                                'b': 42,
                            },
                        },
                    },
                ],
            },
            {
                'role': 'tool',
                'content': 'True',
                'name': 'bar',
            },
            {
                'role': 'tool',
                'content': 'it worked',
                'name': 'foo',
            },
            {
                'role': 'tool',
                'content': 'ghost result',
                'name': 'unknown',
            },
        ]


def test_set_cost_raw():
    orig_message: Message = {
        'role': 'ai',
        'text': 'Hi',
    }
    assert _set_cost_raw(orig_message, _sample_events_streaming_text) == {
        'role': 'ai',
        'text': 'Hi',
        'cost': {
            'input_tokens': 20,
            'output_tokens': 10,
            'total_tokens': 30,
        },
        'raw': _sample_events_streaming_text,
    }
    assert orig_message == {  # <-- orig message should be unmodified!
        'role': 'ai',
        'text': 'Hi',
    }


@pytest.mark.asyncio
async def test_process_stream_streaming_text():
    events = []
    async def callback(event: EventPayload) -> None:
        events.append(event)
    messages = await _process_stream(
        fake_async(_sample_events_streaming_text),
        callback,
    )
    assert events == [
        ('ai', 'text_event', 'Hello'),
        ('ai', 'text_event', '!'),
        ('ai', 'text_event', ' How'),
        ('ai', 'text_event', ' can'),
        ('ai', 'text_event', ' I'),
        ('ai', 'text_event', ' assist'),
        ('ai', 'text_event', ' you'),
        ('ai', 'text_event', ' today'),
        ('ai', 'text_event', '?'),
    ]
    assert messages == [
        {
            'role': 'ai',
            'text': 'Hello! How can I assist you today?',
            'cost': {
                'input_tokens': 20,
                'output_tokens': 10,
                'total_tokens': 30,
            },
            'raw': _sample_events_streaming_text,
        },
    ]


@pytest.mark.asyncio
async def test_process_stream_tool_call():
    events = []
    async def callback(event: EventPayload) -> None:
        events.append(event)
    messages = await _process_stream(
        fake_async(_sample_events_tool_call),
        callback,
    )
    tool_call: ToolCall = {
        'call_id': 'call_0',
        'call_type': 'function',
        'function': {
            'name': 'evaluate_math_expression',
            'arguments': '{"expression": "2.5 * sin(4.5)"}',
        },
    }
    assert events == [
        ('tool_call', 'text_event', 'evaluate_math_expression({"expression": "2.5 * sin(4.5)"})\n'),
        ('tool_call', 'tool_call_event', tool_call),
    ]
    assert messages == [
        {
            'role': 'tool_call',
            'tools': [tool_call],
            'cost': {
                'input_tokens': 100,
                'output_tokens': 29,
                'total_tokens': 129,
            },
            'raw': _sample_events_tool_call,
        },
    ]


@pytest.mark.asyncio
async def test_process_stream_text_and_tool_call():
    _sample_events = [
        *_sample_events_streaming_text,
        *_sample_events_tool_call,
    ]
    events = []
    async def callback(event: EventPayload) -> None:
        events.append(event)
    messages = await _process_stream(
        fake_async(_sample_events),
        callback,
    )
    tool_call: ToolCall = {
        'call_id': 'call_0',
        'call_type': 'function',
        'function': {
            'name': 'evaluate_math_expression',
            'arguments': '{"expression": "2.5 * sin(4.5)"}',
        },
    }
    assert events == [
        ('ai', 'text_event', 'Hello'),
        ('ai', 'text_event', '!'),
        ('ai', 'text_event', ' How'),
        ('ai', 'text_event', ' can'),
        ('ai', 'text_event', ' I'),
        ('ai', 'text_event', ' assist'),
        ('ai', 'text_event', ' you'),
        ('ai', 'text_event', ' today'),
        ('ai', 'text_event', '?'),
        ('tool_call', 'text_event', 'evaluate_math_expression({"expression": "2.5 * sin(4.5)"})\n'),
        ('tool_call', 'tool_call_event', tool_call),
    ]
    assert messages == [
        {
            'role': 'ai',
            'text': 'Hello! How can I assist you today?',
        },
        {
            'role': 'tool_call',
            'tools': [tool_call],
            'cost': {
                'input_tokens': 20 + 100,
                'output_tokens': 10 + 29,
                'total_tokens': 30 + 129,
            },
            'raw': _sample_events,
        },
    ]


@pytest.mark.asyncio
async def test_process_stream_structured_output():
    events = []
    async def callback(event: EventPayload) -> None:
        events.append(event)
    messages = await _process_stream(
        fake_async(_sample_events_structued_output),
        _wrap_event_callback_convert_ai_text_to_tool_call_text(callback),
    )
    assert events == [
        ('tool_call', 'text_event', '{"base":2,"exponent":101}'),
    ]
    assert messages == [
        {
            'role': 'ai',
            'text': '{"base":2,"exponent":101}',
            'cost': {
                'input_tokens': 28,
                'output_tokens': 15,
                'total_tokens': 43,
            },
            'raw': _sample_events_structued_output,
        },
    ]


class MyTypedDict(TypedDict):
    """My special type which is a TypedDict"""
    a: str
    b: int
    c: Literal['one', 'two', 'three']

class MyPydanticType(BaseModel):
    """My special type which is a Pydantic thing"""
    a: str
    b: int
    c: Literal['one', 'two', 'three']
    d: MyTypedDict

def test_get_ollama_format_for_structured_output():
    assert _get_ollama_format_for_structured_output(MyTypedDict) == {
        'type': 'object',
        'title': 'MyTypedDict',
        'description': 'My special type which is a TypedDict',
        'properties': {
            'a': {
                'type': 'string',
                'title': 'A',
            },
            'b': {
                'type': 'integer',
                'title': 'B',
            },
            'c': {
                'type': 'string',
                'title': 'C',
                'enum': ['one', 'two', 'three'],
            },
        },
        'required': ['a', 'b', 'c'],
        'additionalProperties': False,
    }
    assert _get_ollama_format_for_structured_output(MyPydanticType) == {
        'type': 'object',
        'title': 'MyPydanticType',
        'description': 'My special type which is a Pydantic thing',
        'properties': {
            'a': {
                'type': 'string',
                'title': 'A',
            },
            'b': {
                'type': 'integer',
                'title': 'B',
            },
            'c': {
                'type': 'string',
                'title': 'C',
                'enum': ['one', 'two', 'three'],
            },
            'd': {
                '$ref': '#/$defs/MyTypedDict',
            },
        },
        'required': ['a', 'b', 'c', 'd'],
        'additionalProperties': False,
        '$defs': {
            'MyTypedDict': {
                'type': 'object',
                'title': 'MyTypedDict',
                'description': 'My special type which is a TypedDict',
                'properties': {
                    'a': {
                        'type': 'string',
                        'title': 'A',
                    },
                    'b': {
                        'type': 'integer',
                        'title': 'B',
                    },
                    'c': {
                        'type': 'string',
                        'title': 'C',
                        'enum': ['one', 'two', 'three'],
                    },
                },
                'required': ['a', 'b', 'c'],
                'additionalProperties': False,
            },
        },
    }


@pytest.mark.asyncio
async def test_LasagnaOllama_run_once_invalid_arguments():
    async def callback(event: EventPayload) -> None:
        assert event

    with pytest.raises(ValueError, match=r"Oops! You cannot do both tool-use and structured output at the same time!"):
        await LasagnaOllama(model='mistral-small')._run_once(
            event_callback = callback,
            messages = [],
            tools_spec = _convert_to_ollama_tools([sample_tool]),
            force_tool = True,
            format = {'type': 'object'},
        )

    with pytest.raises(ValueError, match=r"Oops! Ollama currently does not support \*optional\* tool use."):
        await LasagnaOllama(model='mistral-small')._run_once(
            event_callback = callback,
            messages = [],
            tools_spec = _convert_to_ollama_tools([sample_tool]),
            force_tool = False,
            format = None,
        )

    with pytest.raises(ValueError, match=r"Oops! You cannot force tools that are not specified!"):
        await LasagnaOllama(model='mistral-small')._run_once(
            event_callback = callback,
            messages = [],
            tools_spec = None,
            force_tool = True,
            format = None,
        )



import pytest

from lasagna.agent_runner import run

from lasagna.types import (
    AgentSpec,
    AgentRun,
    EventCallback,
    Message,
    Model,
    EventPayload,
)

from lasagna.registrar import (
    register_agent,
    register_provider,
    AGENTS,
    PROVIDERS,
)

from lasagna.mock_provider import MockProvider

from typing import List


async def agent_1(
    model: Model,
    event_callback: EventCallback,
    prev_runs: List[AgentRun],
) -> AgentRun:
    assert len(prev_runs) == 0
    messages: List[Message] = []
    new_messages = await model.run(event_callback, messages, [])
    return {
        'agent': 'agent_1',
        'type': 'messages',
        'messages': new_messages,
    }


@pytest.mark.asyncio
async def test_run_with_registered_names():
    AGENTS.clear()
    PROVIDERS.clear()
    register_agent('agent_1', 'Agent 1', agent_1)
    register_provider('mock_provider', 'Mock Provider', MockProvider, [])
    spec: AgentSpec = {
        'agent': 'agent_1',
        'provider': 'mock_provider',
        'model': 'some_model',
        'model_kwargs': {
            'b': 6,
            'a': 'yes',
        },
    }
    events: List[EventPayload] = []
    async def event_callback(event: EventPayload) -> None:
        events.append(event)
    prev_runs: List[AgentRun] = []
    new_run = await run(spec, event_callback, prev_runs)
    assert new_run == {
        'agent': 'agent_1',
        'provider': 'mock_provider',
        'model': 'some_model',
        'model_kwargs': {
            'b': 6,
            'a': 'yes',
        },
        'type': 'messages',
        'messages': [
            {
                'role': 'ai',
                'text': f"model: some_model",
            },
            {
                'role': 'human',
                'text': f"model_kwarg: a = yes",
            },
            {
                'role': 'human',
                'text': f"model_kwarg: b = 6",
            },
        ],
    }
    assert events == [
        ('agent', 'start', 'agent_1'),
        ('ai', 'text_event', 'Hi!'),
        ('agent', 'end', new_run),
    ]


@pytest.mark.asyncio
async def test_run_direct():
    AGENTS.clear()
    PROVIDERS.clear()
    spec: AgentSpec = {
        'agent': agent_1,
        'provider': MockProvider,
        'model': 'some_model',
        'model_kwargs': {
            'b': 6,
            'a': 'yes',
        },
    }
    events: List[EventPayload] = []
    async def event_callback(event: EventPayload) -> None:
        events.append(event)
    prev_runs: List[AgentRun] = []
    new_run = await run(spec, event_callback, prev_runs)
    assert new_run == {
        'agent': 'agent_1',
        'provider': 'MockProvider',
        'model': 'some_model',
        'model_kwargs': {
            'b': 6,
            'a': 'yes',
        },
        'type': 'messages',
        'messages': [
            {
                'role': 'ai',
                'text': f"model: some_model",
            },
            {
                'role': 'human',
                'text': f"model_kwarg: a = yes",
            },
            {
                'role': 'human',
                'text': f"model_kwarg: b = 6",
            },
        ],
    }
    assert events == [
        ('agent', 'start', 'agent_1'),
        ('ai', 'text_event', 'Hi!'),
        ('agent', 'end', new_run),
    ]



from lasagna import (
    known_models,
    flat_messages,
    build_simple_agent,
    extract_last_message,
    noop_callback,
    Message,
)

from typing import List, Dict

import asyncio

from dotenv import load_dotenv; load_dotenv()


THE_AGENT = build_simple_agent(name='agent')


COMMITTEE_SPEC = [
    known_models.BIND_OPENAI_gpt_4o(),
    known_models.BIND_OPENAI_gpt_4o_mini(),

    known_models.BIND_ANTHROPIC_claude_35_sonnet(),
    known_models.BIND_ANTHROPIC_claude_3_opus(),
    known_models.BIND_ANTHROPIC_claude_3_haiku(),

    known_models.BIND_NVIDIA_meta_llama3_8b_instruct(),
    known_models.BIND_NVIDIA_meta_llama3_1_8b_instruct(),
    known_models.BIND_NVIDIA_meta_llama3_2_3b_instruct(),
    known_models.BIND_NVIDIA_mistralai_mistral_large(),
    known_models.BIND_NVIDIA_mistralai_mixtral_8x7b_instruct(),
    known_models.BIND_NVIDIA_google_recurrentgemma_2b(),
    known_models.BIND_NVIDIA_microsoft_phi_3_mini_128k_instruct(),
]


COMMITTEE_MODELS = [
    binder(THE_AGENT)
    for binder in COMMITTEE_SPEC
]


async def vote_on_jokes(joke_a: str, joke_b: str) -> Dict[str, int]:
    messages: List[Message] = [
        {
            'role': 'system',
            'text': "The user will provider you with two jokes named `a` and `b`. You will output which is funniest. Output just a single character (`a` or `b`) to cast your vote.",
        },
        {
            'role': 'human',
            'text': f"a: {joke_a}\n\nb: {joke_b}\n\nRemember, output **only** a single token a or b without any preamble.",
        },
    ]
    tasks = [
        model(noop_callback, [flat_messages('input', messages)])
        for model in COMMITTEE_MODELS
    ]
    outputs = await asyncio.gather(*tasks)
    counter: Dict[str, int] = {}
    for spec, out in zip(COMMITTEE_SPEC, outputs):
        last_message = extract_last_message(out, from_layered_agents=False)
        assert last_message['role'] == 'ai'
        text = last_message['text']
        assert text
        text = text.strip()
        print(spec, text)
        if text not in counter:
            counter[text] = 0
        counter[text] += 1
    print('---------')
    return counter


async def main() -> None:
    #joke_a = "Why don't skeletons fight each other? They don't have the guts."
    #joke_a = "Why did the chicken cross the road? To get to the other side!"
    #joke_a = "I'm not funny."
    joke_a = "Why did the cat jump? Because cats like to jump..."
    joke_b = "Why do programmers prefer dark mode? Because light attracts bugs!"
    votes1 = await vote_on_jokes(joke_a, joke_b)
    votes2 = await vote_on_jokes(joke_b, joke_a)
    votes = {
        'a': (votes1.get('a', 0) + votes2.get('b', 0)),
        'b': (votes1.get('b', 0) + votes2.get('a', 0)),
    }
    print(votes)


if __name__ == '__main__':
    asyncio.run(main())
